;----------------------------------------------------------------------------;
; KEYIRQ.ASM								     ;
;									     ;
; Routines used to convert keyboard presses to flags.			     ;
;									     ;
; Date:    03/07/97							     ;
; Author:  Zonn Moore							     ;
; E-mail:  zonn@concentric.net (as of the above date)			     ;
;									     ;
; Released to the public domain by the author, distribute freely.	     ;
;----------------------------------------------------------------------------;

		model	large

i_8259		equ	020h		; Interrupt controller
i_KEYBD		equ	060h		; Keyboard port

IRQOFS		equ	8		; IRQ offset from system's IRQs
KEYVECT		equ	(IRQOFS+1)	; Vector number of Keyboard IRQ

EOI		equ	020h		; End Of Interrupt

		DataSeg
		public	_KeyFlagTbl

_KeyFlagTbl	db	128 dup (?)	; keyboard flag table

oKeyVect	dd	?		; old keyboard vector
		ends

		CodeSeg
dataSegVal	dw	?		; Segment of data, needed for IRQ

;----------------------------------------------------------------------------;
; Called to Initialize keyboard interrupt vectors to new keyboard handler.   ;
;									     ;
; Assumes:								     ;
;    DS = Data segment							     ;
;									     ;
; Uses:									     ;
;    AX, BX, DX, ES							     ;
;----------------------------------------------------------------------------;
		public	_keyInit
_keyInit	proc
		push	ax
		push	bx
		push	cx
		push	dx
		push	di
		push	es
		

; Save the Data segments value in the Code segment, so that we have access
; to the Data segment at interrupt time.

		mov	cs:[dataSegVal],ds	; save data segment

		mov	di,ds			; move DS to ES
		mov	es,di
		mov	di,Offset _KeyFlagTbl	; point to table
		mov	cx,128			; size of table
		sub	al,al			; zero register
	rep	stosb				; zero all flags

		sub	bx,bx			; point to ABS 0
		mov	es,bx
		mov	bx,KEYVECT*4		; keyboard vector number
		mov	ax,es:[bx]		; get in DX:AX register
		mov	dx,es:[bx+2]
		mov	Word Ptr ds:[oKeyVect],ax ; save old vector
		mov	Word Ptr ds:[oKeyVect+2],dx

		cli				; mask IRQs
		mov	Word Ptr es:[bx],Offset keyIrq ; save new vector
		mov	Word Ptr es:[bx+2],cs
		sti				; restart them

		pop	es
		pop	di
		pop	dx
		pop	cx
		pop	bx
		pop	ax
		ret
_keyInit	endp

;----------------------------------------------------------------------------;
; Restore original keyboard interrupts.					     ;
;									     ;
; Assumes:								     ;
;    DS = Data segment							     ;
;									     ;
; Uses:									     ;
;    AX, BX, DX, ES							     ;
;----------------------------------------------------------------------------;
		public	_keyRestore
_keyRestore	proc
		push	ax
		push	bx
		push	dx
		push	es

		sub	bx,bx			   ; point to ABS 0
		mov	es,bx
		mov	bx,KEYVECT*4		   ; keyboard vector number
		mov	ax,Word Ptr ds:[oKeyVect]  ; get old vector
		mov	dx,Word Ptr ds:[oKeyVect+2]

		cli				   ; mask IRQs
		mov	Word Ptr es:[bx],ax	   ; restore original vector
		mov	Word Ptr es:[bx+2],dx
		sti				   ; restart them

		pop	es
		pop	dx
		pop	bx
		pop	ax
		ret
_keyRestore	endp

;----------------------------------------------------------------------------;
; New keyboard IRQ handler.  Looks up key in scan code table and sets data   ;
; accordingly.								     ;
;----------------------------------------------------------------------------;
keyIrq		proc	far

		sti			; re-enable IRQs
		push	ds
		push	ax
		push	bx

		mov	ds,cs:[dataSegVal] ; point to the Data segment

		in	al,i_KEYBD	; read key
		test	al,080h		; high bit set?
		jnz	resetFlg	; yes, reset key's flag
		mov	bl,al		; use key code as lookup value
		sub	bh,bh		; zero upper byte
		mov	ds:[_KeyFlagTbl+bx],1 ; set key flag
		
		mov	al,EOI		; end of interrupt
		out	i_8259,al	; reset interrupt
		pop	bx
		pop	ax
		pop	ds
		iret

resetFlg:	and	al,07Fh		; remove high bit
		mov	bl,al		; use key code as lookup value
		sub	bh,bh		; zero upper byte
		mov	ds:[_KeyFlagTbl+bx],0 ; reset key flag
		
		mov	al,EOI		; end of interrupt
		out	i_8259,al	; reset interrupt
		pop	bx
		pop	ax
		pop	ds
		iret
keyIrq		endp

		end
