;----------------------------------------------------------------------------;
; Cinemu I/O routines.
;
; Author: Zonn Moore
; Date:   3/16/00
;
; (c) Copyright 2000, Zonn Moore.  All rights reserved.
;----------------------------------------------------------------------------;
		Ideal
		P386N

ADJTMR		equ	0		; 1=Adjust 8254, 0=Don't
SW_ABORT	equ	00100h		; Internally defined Abort Switch

i_8259		equ	020h		; Interrupt controller
i_8254		equ	040h		; Timer
i_KEYBD		equ	060h		; Keyboard port

IRQOFS		equ	8		; IRQ offset from system IRQ
TMRVECT		equ	IRQOFS		; Vector number of Timer IRQ
KEYVECT		equ	(IRQOFS+1)	; Vector number of Keyboard IRQ

EOI		equ	020h		; End Of Interrupt

; Keyboard equates

INMAPSIZ	equ	16		; 16 bytes to map an input
MAXINPUT	equ	64		; maximum number of inputs

KeyMap		equ	(INMAPSIZ * MAXINPUT)	; start key mapping
Inputs		equ	0			; start of inputs

segment		_DATA	public 'DATA' dword
		extrn	_ioInputs:Word,_ioOutputs:Word,_ioSwitches:Byte
		extrn	refreshInc:Word,refreshCnt:Word
		extrn	KeySeg:Word,MKeySeg:Word

 ifdef DEBUGGER
		extrn	_insCount:Word
 endif

oKeyVect	dd	?		; old keyboard vector
oTmrVect	dd	?		; old timer vector
TmrCnt		db	?		; count IRQ timeouts
ScnCnt		db	?		; count 'TmrCnt' timeouts
		ends

;----------------------------------------------------------------------------;
; IOWAIT								     ;
;									     ;
; Macro for wasting time after an I/O access.				     ;
;----------------------------------------------------------------------------;
macro		IOWAIT
		local	carry,nocarry
		%NOCREF	carry,nocarry
		jc	carry	;; give the appearence of real code to any
		jnc	nocarry	;; intelligent instruction chacheing processor.
carry:		stc
nocarry:
endm	

segment		_TEXT	public 'CODE' dword
		assume	cs:_TEXT

		extrn	DataSegment:Word

;----------------------------------------------------------------------------;
; Enables emulation by setting up interrupt and vectors and the such to	     ;
; emulation mode.							     ;
;
; Called with:
;    DS = DataSeg
;									     ;
; Uses:									     ;
;    BX, ES								     ;
;----------------------------------------------------------------------------;
public		hwEnable
proc		hwEnable	near
		push	ax
		push	dx
		call	keyInit		; initialize keyboard routines
		call	timerInit	; setup timer
		pop	dx
		pop	ax
		ret
		endp
		
;----------------------------------------------------------------------------;
; Disables emulation by reseting up interrupt and vectors and the such to    ;
; PC mode.								     ;
;									     ;
; Called with:
;    DS = DataSeg
;									     ;
; Uses:									     ;
;    BX, ES								     ;
;----------------------------------------------------------------------------;
public		hwDisable
proc		hwDisable	near
		push	ax
		push	dx
		call	keyRestore	; restore keyboard to BIOS
		call	timerRestore	; restore timer
		pop	dx
		pop	ax
		ret
		endp
		
;----------------------------------------------------------------------------;
; Initialize keyboard vectors to new key board handler.			     ;
;									     ;
; BIOS routines cannot be called since they might change the FS.	     ;
; registers.								     ;
;									     ;
; Assumes:								     ;
;    DS = Data segment							     ;
;									     ;
; Uses:									     ;
;    AX, BX, DX, ES							     ;
;----------------------------------------------------------------------------;
proc		keyInit		near

		sub	bx,bx			   ; point to ABS 0
		mov	es,bx
		mov	bx,KEYVECT*4		   ; keyboard vector number
		mov	ax,[es:bx]		   ; get in DX:AX register
		mov	dx,[es:bx+2]
		mov	[Word ds:oKeyVect],ax	   ; save old vector
		mov	[Word ds:oKeyVect+2],dx

		cli				   ; mask IRQs
		mov	[Word es:bx],Offset keyIrq ; save new vector
		mov	[Word es:bx+2],cs
		sti				   ; restart them
		ret
		endp

;----------------------------------------------------------------------------;
; Restore original keyboard vectors.					     ;
;									     ;
; BIOS routines cannot be called since they might change the FS.	     ;
; registers.								     ;
;									     ;
; Assumes:								     ;
;    DS = Data segment							     ;
;									     ;
; Uses:									     ;
;    AX, BX, DX, ES							     ;
;----------------------------------------------------------------------------;
proc		keyRestore	near

		sub	bx,bx			   ; point to ABS 0
		mov	es,bx
		mov	bx,KEYVECT*4		   ; keyboard vector number
		mov	ax,[Word ds:oKeyVect]	   ; get old vector
		mov	dx,[Word ds:oKeyVect+2]

		cli				   ; mask IRQs
		mov	[Word es:bx],ax		   ; restore original vector
		mov	[Word es:bx+2],dx
		sti				   ; restart them

		ret
		endp

;----------------------------------------------------------------------------;
; Initialize timer.							     ;
;----------------------------------------------------------------------------;
proc		timerInit	near

 if ADJTMR
		mov	al,034h			; set to mode 2

		cli
		out	043h,al
		IOWAIT
		sub	al,al			; zero count
		out	040h,al			; set counter
		IOWAIT
		out	040h,al
		sti
 endif

		sub	al,al

		cli
		out	i_8254+3,al		; latch timer 0
		IOWAIT				; wait for I/O
		in	al,i_8254		; get timer
		IOWAIT				; wait for I/O
		xchg	al,ah
		in	al,i_8254
		sti

		xchg	al,ah
		neg	ax			; make it an up counter
		mov	[ds:refreshCnt],ax	; save refresh value

		ret

		endp

;----------------------------------------------------------------------------;
; Restore timer.							     ;
;----------------------------------------------------------------------------;
proc		timerRestore	near

 if ADJTMR
		mov	al,36h		; set back to mode 3

		cli
		out	043h,al
		IOWAIT
		sub	al,al		; zero count
		out	040h,al		; set counter
		IOWAIT
		out	040h,al
		sti
 endif

		ret
		endp

;----------------------------------------------------------------------------;
; Check for mouse driver ("C" callable).				     ;
; If found, resets mouse driver.					     ;
;									     ;
; Returns:								     ;
;   0 - Ok, driver found.						     ;
;   1 = Driver not found.						     ;
;----------------------------------------------------------------------------;
public		_checkMouse
proc		_checkMouse	c near

		mov	ax,3533h	; get interupt vector 33h
		int	21h
		mov	ax,es		; get segment
		or	ax,bx		; check for NULL pointer
		jz	noMouse		; mouse not found
		cmp	[Byte es:bx],207 ; check for RTI
		je	noMouse		; mouse not found

		sub	ax,ax		; reset mouse
		int	33h		; call mouse driver
		cmp	ax,-1		; check for driver present
		jne	noMouse		; -1 indicates mouse reset
		sub	ax,ax		; indicate mouse driver exists
		ret

noMouse:	mov	ax,1		; mouse not found
		ret
		endp

;----------------------------------------------------------------------------;
; External Key driver - Press						     ;
;									     ;
; Called to emulate the pressing a key.					     ;
;									     ;
; Called with:								     ;
;    BX = Index into Mouse/Joystick table.				     ;
;----------------------------------------------------------------------------;
public		keyPress
proc		keyPress	near
		push	ax
		push	bx
		mov	es,[ds:MKeySeg]		 ; point to Mouse and Joystick keys
		mov	bl,[es:bx]		 ; get index into INPUT tables
		shl	bx,4			 ; index into 16 byte blocks
		mov	es,[ds:KeySeg]		 ; point to index tables
		mov	ax,[Word es:bx+Inputs]	 ; get low 'OR' mask
		or	[Word ds:_ioInputs],ax	 ; or with low mask
		mov	ax,[Word es:bx+Inputs+2] ; get high 'OR' mask
		or	[Word ds:_ioSwitches],ax ; or with high mask
		mov	ax,[Word es:bx+Inputs+4] ; get low 'AND' mask
		and	[Word ds:_ioInputs],ax	 ; and with low mask
		mov	ax,[Word es:bx+Inputs+6] ; get high 'AND' mask
		and	[Word ds:_ioSwitches],ax ; and with high mask
		pop	bx
		pop	ax
		test	[Word ds:_ioSwitches],SW_ABORT	; abort sw?
		jnz	keyAbort		  ; set up for exit and return
		ret				  ; else, just return
		endp

;----------------------------------------------------------------------------;
; External Key driver - Release						     ;
;									     ;
; Called to emulate the releasing of a key.				     ;
;									     ;
; Called with:								     ;
;    BX = Index into Mouse/Joystick table.				     ;
;----------------------------------------------------------------------------;
public		keyRelease
proc		keyRelease	near
		push	ax
		push	bx
		mov	es,[ds:MKeySeg]		  ; point to Mouse and Joystick keys
		mov	bl,[es:bx]		  ; get index into INPUT tables
		shl	bx,4			  ; index into 16 byte blocks
		mov	es,[ds:KeySeg]		  ; point to index tables
		mov	ax,[Word es:bx+Inputs+8]  ; get low 'OR' mask
		or	[Word ds:_ioInputs],ax	  ; or with low mask
		mov	ax,[Word es:bx+Inputs+10] ; get high 'OR' mask
		or	[Word ds:_ioSwitches],ax  ; or with high mask
		mov	ax,[Word es:bx+Inputs+12] ; get low 'AND' mask
		and	[Word ds:_ioInputs],ax	  ; and with low mask
		mov	ax,[Word es:bx+Inputs+14] ; get high 'AND' mask
		and	[Word ds:_ioSwitches],ax  ; and with high mask
		pop	bx
		pop	ax
		test	[Word ds:_ioSwitches],SW_ABORT	; abort sw?
		jnz	keyAbort		  ; set up for exit and return
		ret				  ; else, just return
		endp

;----------------------------------------------------------------------------;
; Setup opcode table for to abort emulator next time an instruction is	     ;
; decoded.								     ;
;----------------------------------------------------------------------------;
		extrn	opCodeTblA:word,cineExit_A:near

proc		keyAbort	near

 ifdef DEBUGGER
		mov	[ds:_insCount],1	; exit after next instruction
 else

		push	di
		push	cx
		push	cs
		pop	es
		mov	di,Offset opCodeTblA
		mov	ax,Offset cineExit_A
		mov	cx,256 * 4		; number of table entries
	rep	stosw				; all entries point to exit
		pop	cx
		pop	di
		ret
 endif
		endp

;----------------------------------------------------------------------------;
; New keyboard IRQ handler.  Looks up key in scan code table and sets data   ;
; accordingly.								     ;
;----------------------------------------------------------------------------;
proc		keyIrq		near

		sti				; re-enable IRQs
		push	ax
		push	bx
		push	es
		push	ds

		mov	ds,[cs:DataSegment]	; get data segment

		mov	es,[ds:KeySeg]
		in	al,i_KEYBD	; read key

 ifdef DEBUGGER
		cmp	al,001h		; ESC key?
		je	keyIRQabort	; yes, abort
 endif

		mov	bl,al		; get key
		sub	bh,bh		; get word
		test	bl,bl		; key released?
		js	keyHRel		; yes, do key release

; Set memory mapped bit

		mov	bl,[es:bx+KeyMap]	 ; get offset into 'Keys' table
		cmp	bl,0FFh			 ; unused key?
		je	keyDone			 ; nothing to do

		shl	bx,4			 ; point to 16 byte blocks
		mov	ax,[Word es:bx+Inputs]	 ; get low 'OR' mask
		or	[Word ds:_ioInputs],ax	 ; or with low mask
		mov	ax,[Word es:bx+Inputs+2] ; get high 'OR' mask
		or	[Word ds:_ioSwitches],ax ; or with high mask
		mov	ax,[Word es:bx+Inputs+4] ; get low 'AND' mask
		and	[Word ds:_ioInputs],ax	 ; and with low mask
		mov	ax,[Word es:bx+Inputs+6] ; get high 'AND' mask
		and	[Word ds:_ioSwitches],ax ; and with high mask
		test	[Word ds:_ioSwitches],SW_ABORT	; abort sw?
		jnz	keyIRQabort

keyDone:	mov	al,EOI			; end of interrupt
		out	i_8259,al		; reset interrupt
		pop	ds
		pop	es
		pop	bx
		pop	ax
		iret

; Reset memory mapped bit

keyHRel:	and	bl,07Fh			; mask high bit
		mov	bl,[es:bx+KeyMap]	; get offset into 'Keys' table
		cmp	bl,0FFh			; unused key?
		je	keyDone			; nothing to do

		shl	bx,4			  ; point to 16 byte blocks
		mov	ax,[Word es:bx+Inputs+8]  ; get low 'OR' mask
		or	[Word ds:_ioInputs],ax	  ; or with low mask
		mov	ax,[Word es:bx+Inputs+10] ; get high 'OR' mask
		or	[Word ds:_ioSwitches],ax  ; or with high mask
		mov	ax,[Word es:bx+Inputs+12] ; get low 'AND' mask
		and	[Word ds:_ioInputs],ax	  ; and with low mask
		mov	ax,[Word es:bx+Inputs+14] ; get high 'AND' mask
		and	[Word ds:_ioSwitches],ax  ; and with high mask
		test	[Word ds:_ioSwitches],SW_ABORT	; abort sw?
		jnz	keyIRQabort

		mov	al,EOI			; end of interrupt
		out	i_8259,al		; reset interrupt
		pop	ds
		pop	es
		pop	bx
		pop	ax
		iret

keyIRQabort:	call	keyAbort		; do abort 
		mov	al,EOI			; end of interrupt
		out	i_8259,al		; reset interrupt
		pop	ds
		pop	es
		pop	bx
		pop	ax
		iret
		endp

 if 0
;----------------------------------------------------------------------------;
; Read joystick's value							     ;
;----------------------------------------------------------------------------;
proc		readJoys

; AH  = Mask of values to ignore.
; SI  = Pointer to array of values.
; ECX = Counter.

readLoop:	in	al,dx			; get current values
		test	al,ah			; any changes?
		jz	readL1			; no, skip
		mov	[ds:si],al		; save read values
		mov	[ds:si+1],ecx		; save count
		add	si,5			; skip reading
		jnz	readL2			; branch always taken

readL1:		mov	[ds:si],al		; dummy save
		mov	[ds:si+1],ecx		; dummy save
		add	si,0			; dummy add
		jz	readLoop		; dummy, branch never taken
readL2:		not	al			; set read bits low
		and	ah,al			; mask them from being tested
		jz	readDone		; if no more bit left, then done
		inc	ecx			; increment pointer
		jnz	readLoop		; if no overflow, read again

; If it ever falls through, then an overflow error has occured.

readErr:	ret
 endif

;----------------------------------------------------------------------------;
; Some general housekeeping routines					     ;
;----------------------------------------------------------------------------;

;----------------------------------------------------------------------------;
; uchar far *paraAdj( uchar far *aPtr)					     ;
;									     ;
; Adjust 'far' memory pointer to the next paragraph (modulo 16) boundry.     ;
; If already pointing to a paragraph, then no adjustment will be made.	     ;
;									     ;
; Called with:								     ;
;    ptr = Pointer to far memory.					     ;
;									     ;
; Returns:								     ;
;    Far memory pointer on paragraph boundry.				     ;
;----------------------------------------------------------------------------;
public		_paraAdj
proc		_paraAdj	c

		arg	aPtr:Dword

		mov	ax,[Word aPtr]	; get pointer in DX:AX
		mov	dx,[Word aPtr+2]
		shr	ax,1		; convert to segment
		shr	ax,1
		shr	ax,1
		shr	ax,1
		add	dx,ax		; add number of paragraphs in offset
		mov	ax,[Word aPtr]	; get original offset
		and	al,00Fh		; save only paragraph remainder
		add	al,0FFh		; was offset 0?
		adc	dx,0		; no, increment segment
		sub	ax,ax		; return pointer in DX:AX
		ret
		endp

;----------------------------------------------------------------------------;
; vsetmode( mode)							     ;
;									     ;
; Set current video mode.						     ;
;									     ;
; Where:								     ;
;    mode = Mode to set video.						     ;
;    	0 =  40 x  25 b&w.						     ;
;    	1 =  40 x  25 color.						     ;
;    	2 =  80 x  25 b&w.						     ;
;    	3 =  80 x  25 color.						     ;
;    	4 = 320 x 200 color.						     ;
;    	5 = 320 x 200 b&w.						     ;
;    	6 = 640 x 200 b&w.						     ;
;    	7 = Mono card.							     ;
;----------------------------------------------------------------------------;
public		_vsetmode
proc		_vsetmode	c

		arg	mode:Byte

		mov	al,[mode]	; get mode
		mov	ah,0		; Set Mode
		int	10h

		mov	dx,01CEh
		mov	al,0B0h
		cli
		out	dx,al
		inc	dx
		in	al,dx
		dec	dx
		mov	ah,al
		mov	al,0B0h
		out	dx,al
		inc	dx
		mov	al,ah
		and	al,0BFh
		out	dx,al
		sti
		
		ret
		endp

		ends
		end
