;----------------------------------------------------------------------------;
; Cinematronic's CPU emulator.
;
; Author: Zonn Moore
; Date:   3/16/00
;
; (c) Copyright 2000, Zonn Moore.  All rights reserved.
;----------------------------------------------------------------------------;
		Ideal
		P386N
		include	"zvga.inc"

;DEBUGGER	equ	0		; if defined, use debugger
;LOG_IO_HITS	equ	0

 ifndef DEBUGGER
MACRORET	equ	1		; if defined, use macros
 endif

TRACE		equ	0
WRITEADR	equ	0
BKPOINTS	equ	0

DISBRK		equ	0FFFFh		; break disabled

STATE_A		equ	0
STATE_AA	equ	1
STATE_B		equ	2
STATE_BB	equ	3

B_FF		equ	080h		; Use B-reg Flip Flop flag
ACC_F		equ	07Fh		; indicate last inst was a ACC inst
OP_MUL1		equ	0E3h		; opCode for multiply
OP_MUL2		equ	0F3h		; opCode for multiply

i_8254		equ	040h		; Timer

; Switch definitions

SW_ABORT	equ	00100h		; Internally defined Abort Switch
SW_RESET	equ	00200h		; Internally defined Reset Switch
SW_COIN		equ	00080h		; Coin Switch

segment		_DATA	public 'DATA' dword
		extrn	_vgSeg:Word,vgVecPtr:Dword

		public	_ramBank,_ioInputs,_ioSwitches,_ioOutputs
		public	_accVal,_cmpVal
		public	_pa0Cf,_pa0,_cFlag

		public	_eRegPC,_eRegA,_eRegB,_eRegI,_eRegJ,_eRegP,_eCState

		public	KeySeg,refreshCnt
		public	MouseType,MKeySeg,OMouseKeys,VMouseKeys
		public	JoyX,JoyY,MouseSpeedX,MouseSpeedY

; <<<<<<<<<<< The following item between arrows are the CPU state >>>>>>>>>>>
; <<<<<<<<<<< variables, they are order dependent >>>>>>>>>>>>>>>>>>>>>>>>>>>

_ramBank	dw	256 dup(?)	; RAM registers

; CPU flags and temp storage

_accVal		dw	?		; last ACC
_cmpVal		dw	?		; new Acc

; Do not change order of next three labels

label		_pa0Cf	word
_pa0		db	?		; Bit zero of A-reg at last ACC access
_cFlag		db	?		; carry flag indicator

; Registers

_eRegPC		dw	?		; PC-reg
_eRegA		dw	?		; A-reg
_eRegB		dw	?		; B-reg
_eRegI		dw	?		; I-reg
_eRegJ		dw	?		; J-reg
_eRegP		db	?		; P-reg
_eCState	db	?		; current state

; <<<<<<<<<<<<<< End of CPU state variables >>>>>>>>>>>>>>>>>>>>>>>>>>>>>

; Segment addresses

EmuSeg		dw	?		; segment of emulator
KeySeg		dw	?		; segment of key mappings
MKeySeg		dw	?		; segment of Mouse/Joystick key mappings

; 8253 counter values to control refresh rate

refreshRate	dw	?		; rate count of refresh
refreshCnt	dw	?		; current refresh counter

; X and Y Joystick values, as read from DAC

JoyX		dw	?		; integer part
		db	?		; fractional part

JoyY		dw	?		; integer part
		db	?		; fractional part

MouseSpeedX	dw	?		; Scale Mickeys
MouseSpeedY	dw	?		; Scale Mickeys

; Inputs and switches

_ioInputs	dw	?		; input values
_ioSwitches	dw	?		; input switches
_ioOutputs	db	?		; latch outputs

; Mouse buttons at last poll

MouseType	db	?		; type of mouse interface
OMouseKeys	db	?		; previous value of mouse keys
VMouseKeys	db	?		; valid mouse keys

 ifdef DEBUGGER

; Debugger data

public		_insCount
public		_eMiFlg,_eLtFlg,_eEqFlg,_eCyFlg,_eA0Flg,_eBfFlg
public		_oRegPC,_oRegA,_oRegB,_oRegI,_oRegJ,_oRegP
public		_oCState,_oMiFlg,_oLtFlg,_oEqFlg,_oCyFlg,_oA0Flg
public		_oBfFlg,_eInsBrk0,_eInsBrk1,_eInsBrk2,_eInsBrk3
public		_eInsGo

_insCount	dw	?		; instruction counter

_eMiFlg		db	?		; minus flag
_eLtFlg		db	?		; less than flag
_eEqFlg		db	?		; equal flag
_eCyFlg		db	?		; carry flag
_eA0Flg		db	?		; PA0 flag
_eBfFlg		db	?		; B flip/flop flag

_oRegPC		dw	?		; PC-reg
_oRegA		dw	?		; A-reg
_oRegB		dw	?		; B-reg
_oRegI		dw	?		; I-reg
_oRegJ		dw	?		; J-reg
_oRegP		db	?		; P-reg
_oCState	db	?		; current state
_oMiFlg		db	?		; minus flag
_oLtFlg		db	?		; less than flag
_oEqFlg		db	?		; equal flag
_oCyFlg		db	?		; carry flag
_oA0Flg		db	?		; PA0 flag
_oBfFlg		db	?		; B flip/flop flag

_eInsBrk0	dw	?		; break address 0
_eInsBrk1	dw	?		; break address 1
_eInsBrk2	dw	?		; break address 2
_eInsBrk3	dw	?		; break address 3
_eInsGo		dw	?		; go break address

  if TRACE

public		_tracePtr,_traceBfr

_tracePtr	dw	?		; trace pointer
_traceBfr	dw	16 dup (?)	; trace buffer
  endif

  if WRITEADR

public		_writeAdr

_writeAdr	dw	?		; write breakpoint address
  endif
 endif
		ends

;----------------------------------------------------------------------------;
; IOWAIT								     ;
;									     ;
; Macro for wasting time after an I/O access.				     ;
;----------------------------------------------------------------------------;
macro		IOWAIT
		local	carry,nocarry
		%NOCREF	carry,nocarry
		jc	carry	;; give the appearence of real code to any
		jnc	nocarry	;; intelligent instruction chacheing processor.
carry:		stc
nocarry:
endm

macro		SYNC4
		align	4
		endm

macro		SaveRegs
		push	ax
		push	cx
		push	dx
		push	bp
		push	si
		push	di
		push	ds
		push	fs
		endm

macro		RestoreRegs
		pop	fs
		pop	ds
		pop	di
		pop	si
		pop	bp
		pop	dx
		pop	cx
		pop	ax
		endm

segment		_TEXT	public 'CODE' dword
		assume	cs:_TEXT

		extrn	hwEnable:Near,hwDisable:Near

		extrn	vgStart:Near,vgLength:Near,vgDraw:Near,vgSwap:Near
		extrn	vgBrightBI:Near,vgBright16:Near,vgBright64:Near
		extrn	vgBrightCo:Near

		extrn	soundSet:Near,soundReset:Near,soundIO:Near

		extrn	_logIO:Near

		public		DataSegment
DataSegment	dw	?	; must be kept in Code segment for IRQ routines

;----------------------------------------------------------------------------;
; Reset CPU								     ;
;----------------------------------------------------------------------------;
public		_cineReset
proc		_cineReset	c near

		sub	ax,ax

		mov	[ds:_accVal],ax
		mov	[ds:_cmpVal],ax
		mov	[ds:_pa0Cf],ax

		mov	[ds:_eRegPC],ax			; zero PC-reg
		mov	[ds:_eRegA],ax			; zero A-reg
		mov	[ds:_eRegB],ax			; zero B-reg
		mov	[ds:_eRegI],Offset _ramBank	; zero I-reg
		mov	[ds:_eRegJ],ax			; zero J-reg
		mov	[ds:_eRegP],al			; zero P-reg
		mov	[ds:_eCState],STATE_A		; reset state

		; Reset Joystick values

		mov	[ds:JoyX],ax
		mov	[ds:JoyY],ax

 ifdef DEBUGGER
		mov	[ds:_eMiFlg],al			; minus flag
		mov	[ds:_eLtFlg],al			; less than flag
		mov	[ds:_eEqFlg],1			; equal flag
		mov	[ds:_eCyFlg],al			; carry flag
		mov	[ds:_eA0Flg],al			; PA0 flag
		mov	[ds:_eBfFlg],al			; B flip flop flag

		mov	[ds:_oRegPC],ax			; zero PC-reg
		mov	[ds:_oRegA],ax			; zero A-reg
		mov	[ds:_oRegB],ax			; zero B-reg
		mov	[ds:_oRegI],Offset _ramBank	; zero I-reg
		mov	[ds:_oRegJ],ax			; zero J-reg
		mov	[ds:_oRegP],al			; zero P-reg
		mov	[ds:_oCState],STATE_A		; reset state

		mov	[ds:_oMiFlg],al			; minus flag
		mov	[ds:_oLtFlg],al			; less than flag
		mov	[ds:_oEqFlg],1			; equal flag
		mov	[ds:_oCyFlg],al			; carry flag
		mov	[ds:_oA0Flg],al			; PA0 flag
		mov	[ds:_oBfFlg],al			; B flip flop flag

  if TRACE
		mov	[ds:_tracePtr],ax			; reset trace pointer
  endif

  if WRITEADR
		mov	[ds:_writeAdr],ax
  endif
 endif

		ret
		endp

public		_cineSetJMI
proc		_cineSetJMI	c near
		arg	aJMIflag:word

		cmp	[aJMIflag],0		; flag set?
		jnz	setJMI1			; no, setup for JMI
		mov	[cs:tJMI_A_B],Offset opJEI_A_B
		mov	[cs:tJMI_A_A],Offset opJEI_A_A
		mov	[cs:tJMI_AA_B],Offset opJEI_AA_B
		mov	[cs:tJMI_AA_A],Offset opJEI_AA_A
		mov	[cs:tJMI_B_BB1],Offset opJEI_B_BB
		mov	[cs:tJMI_B_BB2],Offset opJEI_B_BB
		mov	[cs:tJMI_BB_B],Offset opJEI_BB_B
		mov	[cs:tJMI_BB_A],Offset opJEI_BB_A
		ret

setJMI1:	mov	[cs:tJMI_A_B],Offset opJMI_A_B
		mov	[cs:tJMI_A_A],Offset opJMI_A_A
		mov	[cs:tJMI_AA_B],Offset opJMI_AA_B
		mov	[cs:tJMI_AA_A],Offset opJMI_AA_A
		mov	[cs:tJMI_B_BB1],Offset opJMI_B_BB
		mov	[cs:tJMI_B_BB2],Offset opJMI_B_BB
		mov	[cs:tJMI_BB_B],Offset opJMI_BB_B
		mov	[cs:tJMI_BB_A],Offset opJMI_BB_A
		ret
		endp

public		_cineSetMSize
proc		_cineSetMSize	c near
		arg	aMemSizeFlag:word

		cmp	[aMemSizeFlag],2	; 16k flag set?
		je	setMSz1			; yes, setup for 16k
		cmp	[aMemSizeFlag],3	; 32k flag set?
		je	setMSz2			; yes, setup for 32k

		; setup for 4k or 8k

		mov	[cs:tJPP_A_B],Offset opJPP8_A_B
		mov	[cs:tJPP_AA_B],Offset opJPP8_AA_B
		mov	[cs:tJPP_B_BB],Offset opJPP8_B_BB
		mov	[cs:tJPP_BB_B],Offset opJPP8_BB_B
		ret

		; setup for 16k

setMSz1:	mov	[cs:tJPP_A_B],Offset opJPP16_A_B
		mov	[cs:tJPP_AA_B],Offset opJPP16_AA_B
		mov	[cs:tJPP_B_BB],Offset opJPP16_B_BB
		mov	[cs:tJPP_BB_B],Offset opJPP16_BB_B
		ret

		; setup for 32k

setMSz2:	mov	[cs:tJPP_A_B],Offset opJPP32_A_B
		mov	[cs:tJPP_AA_B],Offset opJPP32_AA_B
		mov	[cs:tJPP_B_BB],Offset opJPP32_B_BB
		mov	[cs:tJPP_BB_B],Offset opJPP32_BB_B
		ret
		endp

public		_cineSetRate
proc		_cineSetRate	c near
		arg	aRate:word

		mov	ax,[aRate]
		mov	[ds:refreshRate],ax
		ret
		endp

public		_cineSetSw
proc		_cineSetSw	c near
		arg	aSwitches:word,aInputs:word

		mov	ax,[aSwitches]		; get switch settings
		mov	[ds:_ioSwitches],ax	; set switch values
		mov	ax,[aInputs]		; get init settings
		mov	[ds:_ioInputs],ax	; set input values
		ret
		endp

public		_cineSetMonitor
proc		_cineSetMonitor	c near
		arg	aMonitor:word
		uses	di

		push	cs
		pop	es

		cmp	[aMonitor],1		; 16-Level?
		je	cineSet16		; yes, set for 16-level
		cmp	[aMonitor],2		; 64-Level?
		je	cineSet64		; yes, set for 64-level
		cmp	[aMonitor],3		; WotW-Color?
		je	cineSetWW		; yes, set of WotW

		; Assume Bi-Level

		mov	ax,Offset opOUTbi_A_A
		mov	di,Offset tOUT_A_A
		mov	cx,16
	rep	stosw

		mov	ax,Offset opOUTbi_AA_A
		mov	di,Offset tOUT_AA_A
		mov	cx,16
	rep	stosw

		mov	ax,Offset opOUTbi_B_BB
		mov	di,Offset tOUT_B_BB
		mov	cx,16
	rep	stosw

		mov	ax,Offset opOUTbi_BB_A
		mov	di,Offset tOUT_BB_A
		mov	cx,16
	rep	stosw
		ret

		; 16-Level routines

cineSet16:	mov	ax,Offset opOUT16_A_A
		mov	di,Offset tOUT_A_A
		mov	cx,16
	rep	stosw

		mov	ax,Offset opOUT16_AA_A
		mov	di,Offset tOUT_AA_A
		mov	cx,16
	rep	stosw

		mov	ax,Offset opOUT16_B_BB
		mov	di,Offset tOUT_B_BB
		mov	cx,16
	rep	stosw

		mov	ax,Offset opOUT16_BB_A
		mov	di,Offset tOUT_BB_A
		mov	cx,16
	rep	stosw
		ret

		; 64-Level routines

cineSet64:	mov	ax,Offset opOUT64_A_A
		mov	di,Offset tOUT_A_A
		mov	cx,16
	rep	stosw

		mov	ax,Offset opOUT64_AA_A
		mov	di,Offset tOUT_AA_A
		mov	cx,16
	rep	stosw

		mov	ax,Offset opOUT64_B_BB
		mov	di,Offset tOUT_B_BB
		mov	cx,16
	rep	stosw

		mov	ax,Offset opOUT64_BB_A
		mov	di,Offset tOUT_BB_A
		mov	cx,16
	rep	stosw
		ret

		; WotW & Boxing Bugs Color

cineSetWW:	mov	ax,Offset opOUTWW_A_A
		mov	di,Offset tOUT_A_A
		mov	cx,16
	rep	stosw

		mov	ax,Offset opOUTWW_AA_A
		mov	di,Offset tOUT_AA_A
		mov	cx,16
	rep	stosw

		mov	ax,Offset opOUTWW_B_BB
		mov	di,Offset tOUT_B_BB
		mov	cx,16
	rep	stosw

		mov	ax,Offset opOUTWW_BB_A
		mov	di,Offset tOUT_BB_A
		mov	cx,16
	rep	stosw
		ret

		endp

;----------------------------------------------------------------------------;
; Setup mouse functions							     ;
;----------------------------------------------------------------------------;
public		_cineSetMouse
proc		_cineSetMouse	c near

		arg	aMouseType:byte,aMouseSpeedX:word,aMouseSpeedY:word
		arg	aKeyMask:byte,aMKeySeg:dword

		mov	ax,[aMouseSpeedX]	; get scaling
		mov	[ds:MouseSpeedX],ax	; save scaling
		mov	ax,[aMouseSpeedY]	; get scaling
		mov	[ds:MouseSpeedY],ax	; save scaling
		mov	ax,[Word aMKeySeg+2]	; get segment
		mov	[ds:MKeySeg],ax		; save mouse key map segment
		mov	al,[aMouseType]		; get type
		mov	[ds:MouseType],al	; save type
		mov	al,[aKeyMask]		; get mask of valid keys
		mov	[ds:VMouseKeys],al	; save mask
		mov	[ds:OMouseKeys],0	; no keys pressed
		ret

		endp

public		_cineInit
proc		_cineInit	c near

		arg	aEmuSeg:dword,aKeyMap:dword

		mov	ax,[Word aEmuSeg+2]	; get segment
		mov	[ds:EmuSeg],ax		; save segment address

		mov	ax,[Word aKeyMap+2]	; get segment
		mov	[ds:KeySeg],ax		; save segment address

 ifdef DEBUGGER
		mov	[ds:_eInsBrk0],DISBRK	; clear breakpoints
		mov	[ds:_eInsBrk1],DISBRK
		mov	[ds:_eInsBrk2],DISBRK
		mov	[ds:_eInsBrk3],DISBRK
		mov	[ds:_eInsGo],DISBRK
 endif

		ret
		endp

 ifdef DEBUGGER

;----------------------------------------------------------------------------;
; This code is used when the emulator is being run from the debugger.	     ;
;----------------------------------------------------------------------------;
public		_cineExec
proc		_cineExec	c near

		arg	count:Word

		push	si
		push	di
		push	bp
		push	ds
		mov	[cs:DataSegment],ds	; save emulator's data seg
		mov	fs,[ds:EmuSeg]		; get emulator's segment
		mov	ax,[count]		; get instruction count
		mov	[ds:_insCount],ax	; save count

		call	hwEnable		; enable hardware for C-CPU

		mov	al,[ds:_eMiFlg]		; reset old flags
		mov	[ds:_oMiFlg],al
		mov	al,[ds:_eLtFlg]
		mov	[ds:_oLtFlg],al
		mov	al,[ds:_eEqFlg]
		mov	[ds:_oEqFlg],al
		mov	al,[ds:_eCyFlg]
		mov	[ds:_oCyFlg],al
		mov	al,[ds:_eA0Flg]
		mov	[ds:_oA0Flg],al
		mov	al,[ds:_eBfFlg]
		mov	[ds:_oBfFlg],al

		; get current values.

		mov	si,[ds:_eRegPC]		; get PC-reg
		mov	ax,[ds:_eRegA]		; get A-reg
		mov	dx,[ds:_eRegB]		; get B-reg
		mov	di,[ds:_eRegI]		; get I-reg
		mov	bp,[ds:_eRegJ]		; get J-reg
		mov	ch,[ds:_eRegP]		; get P-reg
		mov	bl,[ds:_eCState]	; get state

		; save old values to allow debugger to highlight changed
		; registers

		mov	[ds:_oRegPC],si		; set PC-reg
		mov	[ds:_oRegA],ax		; set A-reg
		mov	[ds:_oRegB],dx		; set B-reg
		mov	[ds:_oRegI],di		; set I-reg
		mov	[ds:_oRegJ],bp		; set J-reg
		mov	[ds:_oRegP],ch		; set P-reg
		mov	[ds:_oCState],bl	; set old state

		; reset abort and reset bits

		and	[ds:_ioSwitches],Not (SW_ABORT or SW_RESET)

		; jump to proper state

		test	bl,bl
		jz	cineExec_A
		dec	bl
		jz	cineExec_AA
		dec	bl
		jz	cineExec_B
		jmp	cineExec_BB

		; All states return here

label		cineExit	near

		mov	[ds:_eRegPC],si		; save PC-reg
		mov	[ds:_eRegA],ax		; save A-reg
		mov	[ds:_eRegB],dx		; save B-reg
		mov	[ds:_eRegI],di		; save I-reg
		mov	[ds:_eRegJ],bp		; save J-reg
		mov	[ds:_eRegP],ch		; save P-reg

		sub	al,al			; zero flag
		cmp	[ds:_eCState],STATE_B	; B flip/flop set?
		jne	cineX1			; no, continue
		inc	al			; else, set B_FF flag
cineX1:		mov	[ds:_eBfFlg],al

		; set minus flag based on STATE

		cmp	[ds:_eCState],STATE_AA
		je	cineSAA			; use last Acc
		cmp	[ds:_eCState],STATE_BB
		je	cineSBB			; use B-reg

		mov	al,[Byte ds:_eRegA+1]	; else, get A-reg's value
		jmp	Short cineX2

cineSAA:	mov	al,[Byte ds:_accVal+1]	; get last Acc's value
		jmp	Short cineX2

cineSBB:	mov	al,[Byte ds:_eRegB+1]	; get B-reg's value

cineX2:		shr	al,3			; save just sign flag
		mov	[ds:_eMiFlg],al		; save sign flag

		; get less than flag

		sub	al,al
		mov	bx,[ds:_cmpVal]
		cmp	bx,[ds:_accVal]
		jae	cineX3			; if >= then no flag
		inc	al			; else, set flag
cineX3:		mov	[ds:_eLtFlg],al

		; get equal flag

		sub	al,al
		mov	bx,[ds:_cmpVal]
		cmp	bx,[ds:_accVal]
		jne	cineX4			; if <> then no flag
		inc	al			; else, set flag
cineX4:		mov	[ds:_eEqFlg],al

		; get carry flag

		sub	al,al
		mov	bl,[ds:_cFlag]
		and	bl,0F0h			; any overflow?
		jz	cineX5			; no, then no flag
		inc	al			; else, set flag
cineX5:		mov	[ds:_eCyFlg],al

		; get PA0 flag

		sub	al,al
		mov	bl,[ds:_pa0]
		and	bl,001h			; bit set?
		jz	cineX6			; no, then no flag
		inc	al			; else, set flag
cineX6:		mov	[ds:_eA0Flg],al

		call	hwDisable		; disable C-CPU hardware

		pop	ds
		pop	bp
		pop	di
		pop	si

		sub	ax,ax
		test	[ds:_ioSwitches],SW_RESET
		jz	cineX7
		inc	ax			; indicate reset desired
cineX7:		ret
		endp
 else

;----------------------------------------------------------------------------;
; This code is used when the emulator is NOT being run from the debugger.    ;
;----------------------------------------------------------------------------;
public		_cineExec
proc		_cineExec	c near

		push	si
		push	di
		push	bp
		push	ds

		mov	[cs:DataSegment],ds	; save emulator's data seg
		mov	fs,[ds:EmuSeg]		; get emulator's segment
		push	ds
		pop	gs			; get data segment for IRQs and Video

		call	hwEnable		; enable hardware for C-CPU

		mov	si,[ds:_eRegPC]		; get PC-reg
		mov	ax,[ds:_eRegA]		; get A-reg
		mov	dx,[ds:_eRegB]		; get B-reg
		mov	di,[ds:_eRegI]		; get I-reg
		mov	bp,[ds:_eRegJ]		; get J-reg
		mov	ch,[ds:_eRegP]		; get P-reg
		mov	bl,[ds:_eCState]	; get state

		; reset abort and reset bits

		and	[ds:_ioSwitches],Not (SW_ABORT or SW_RESET)

		; jump to proper state

		test	bl,bl
		jz	cineExec_A
		dec	bl
		jz	cineExec_AA
		dec	bl
		jz	cineExec_B
		jmp	cineExec_BB

; Abort and reset returns here

label		cineExit	near

		mov	[ds:_eRegPC],si		; save PC-reg
		mov	[ds:_eRegA],ax		; save A-reg
		mov	[ds:_eRegB],dx		; save B-reg
		mov	[ds:_eRegI],di		; save I-reg
		mov	[ds:_eRegJ],bp		; save J-reg
		mov	[ds:_eRegP],ch		; save P-reg

		call	hwDisable		; disable C-CPU hardware

		pop	ds
		pop	bp
		pop	di
		pop	si

		sub	ax,ax
		test	[ds:_ioSwitches],SW_RESET
		jz	cineX7
		inc	ax			; indicate reset desired
cineX7:		ret
		endp
 endif

;----------------------------------------------------------------------------;
; The register mapping is as follows:					     ;
;									     ;
;    AX = Cine' A-reg							     ;
;    DX = Cine' B-reg							     ;
;    CH = Cine' P-reg (low nibble, high nibble always zero)		     ;
;    DI = Cine' I-reg							     ;
;    BP = Cine' J-reg							     ;
;    SI = Cine' Address register					     ;
;									     ;
; General use registers:						     ;
;    ES, BX, CL								     ;
;----------------------------------------------------------------------------;

 ifndef MACRORET

; To allow for breakpoints in the debugger, the end of each instruction
; jumps to a routine that does the breakpoint comparisons.

macro		jumpCineRet_A
		jmp	cineRet_A
		endm
		
macro		jumpCineRet_B
		jmp	cineRet_B
		endm

macro		jumpCineRet_AA
		jmp	cineRet_AA
		endm
		
macro		jumpCineRet_BB
		jmp	cineRet_BB
		endm

 else

; If no debugger, each instruction will jump to the start of the next one.

macro		jumpCineRet_A
		sub	bh,bh
		mov	bl,[fs:si]		; get opcode
		add	bx,bx			; index word table
		jmp	[cs:bx+opCodeTblA]	; jump to opcode routine
		endm

macro		jumpCineRet_B
		sub	bh,bh
		mov	bl,[fs:si]		; get opcode
		add	bx,bx			; index word table
		jmp	[cs:bx+opCodeTblB]	; jump to opcode routine
		endm

macro		jumpCineRet_AA
		sub	bh,bh
		mov	bl,[fs:si]		; get opcode
		add	bx,bx			; index word table
		jmp	[cs:bx+opCodeTblAA]	; jump to opcode routine
		endm

macro		jumpCineRet_BB
		sub	bh,bh
		mov	bl,[fs:si]		; get opcode
		add	bx,bx			; index word table
		jmp	[cs:bx+opCodeTblBB]	; jump to opcode routine
		endm
 endif
		
proc		cineMainLoop	near

;----------------------------------------------------------------------------;
; 'A' table								     ;
;----------------------------------------------------------------------------;
		SYNC4
label		cineRet_A	near

 ifdef DEBUGGER
		cmp	[ds:_insCount],0		; no counting?
		je	cineRA1
		dec	[ds:_insCount]
		jz	cineExit_A
cineRA1:
  if BKPOINTS
		cmp	si,[ds:_eInsBrk0]
		je	cineExit_A
		cmp	si,[ds:_eInsBrk1]
		je	cineExit_A
		cmp	si,[ds:_eInsBrk2]
		je	cineExit_A
		cmp	si,[ds:_eInsBrk3]
		je	cineExit_A
		cmp	si,[ds:_eInsGo]
		je	cineExit_A
  endif
 endif

label		cineExec_A	near

 ifdef DEBUGGER
  if TRACE
		mov	bx,[ds:_tracePtr]
		add	bx,2
		and	bx,0001Fh
		mov	[ds:bx+_traceBfr],si
		mov	[ds:_tracePtr],bx
  endif
 endif

		sub	bh,bh
		mov	bl,[fs:si]		; get opcode
		add	bx,bx			; index word table
		jmp	[cs:bx+opCodeTblA]	; jump to opcode routine

; Exit back to Debugger

public		cineExit_A
cineExit_A:	mov	[ds:_eCState],STATE_A
		jmp	cineExit

;----------------------------------------------------------------------------;
; 'B' table								     ;
;----------------------------------------------------------------------------;
		SYNC4
label		cineRet_B	near

 ifdef DEBUGGER
		cmp	[ds:_insCount],0		; no counting?
		je	cineRB1
		dec	[ds:_insCount]
		jz	cineExit_B
cineRB1:
  if BKPOINTS
		cmp	si,[ds:_eInsBrk0]
		je	cineExit_B
		cmp	si,[ds:_eInsBrk1]
		je	cineExit_B
		cmp	si,[ds:_eInsBrk2]
		je	cineExit_B
		cmp	si,[ds:_eInsBrk3]
		je	cineExit_B
		cmp	si,[ds:_eInsGo]
		je	cineExit_B
  endif
 endif

label		cineExec_B	near

 ifdef DEBUGGER
  if TRACE
		mov	bx,[ds:_tracePtr]
		add	bx,2
		and	bx,0001Fh
		mov	[ds:bx+_traceBfr],si
		mov	[ds:_tracePtr],bx
  endif
 endif

		sub	bh,bh
		mov	bl,[fs:si]		; get opcode
		add	bx,bx			; index word table
		jmp	[cs:bx+opCodeTblB]	; jump to opcode routine

; Exit back to Debugger

cineExit_B:	mov	[ds:_eCState],STATE_B
		jmp	cineExit

;----------------------------------------------------------------------------;
; 'AA' table								     ;
;----------------------------------------------------------------------------;
		SYNC4
label		cineRet_AA	near

 ifdef DEBUGGER
		cmp	[ds:_insCount],0		; no counting?
		je	cineRAA1
		dec	[ds:_insCount]
		jz	cineExit_AA
cineRAA1:
  if BKPOINTS
		cmp	si,[ds:_eInsBrk0]
		je	cineExit_AA
		cmp	si,[ds:_eInsBrk1]
		je	cineExit_AA
		cmp	si,[ds:_eInsBrk2]
		je	cineExit_AA
		cmp	si,[ds:_eInsBrk3]
		je	cineExit_AA
		cmp	si,[ds:_eInsGo]
		je	cineExit_AA
  endif
 endif

label		cineExec_AA	near

 ifdef DEBUGGER
  if TRACE
		mov	bx,[ds:_tracePtr]
		add	bx,2
		and	bx,0001Fh
		mov	[ds:bx+_traceBfr],si
		mov	[ds:_tracePtr],bx
  endif
 endif

		sub	bh,bh
		mov	bl,[fs:si]		; get opcode
		add	bx,bx			; index word table
		jmp	[cs:bx+opCodeTblAA]	; jump to opcode routine

; Exit back to Debugger

cineExit_AA:	mov	[ds:_eCState],STATE_AA
		jmp	cineExit

;----------------------------------------------------------------------------;
; 'BB' table								     ;
;----------------------------------------------------------------------------;
		SYNC4
label		cineRet_BB	near

 ifdef DEBUGGER
		cmp	[ds:_insCount],0		; no counting?
		je	cineRBB1
		dec	[ds:_insCount]
		jz	cineExit_BB
cineRBB1:
  if BKPOINTS
		cmp	si,[ds:_eInsBrk0]
		je	cineExit_BB
		cmp	si,[ds:_eInsBrk1]
		je	cineExit_BB
		cmp	si,[ds:_eInsBrk2]
		je	cineExit_BB
		cmp	si,[ds:_eInsBrk3]
		je	cineExit_BB
		cmp	si,[ds:_eInsGo]
		je	cineExit_BB
  endif
 endif

label		cineExec_BB	near

 ifdef DEBUGGER
  if TRACE
		mov	bx,[ds:_tracePtr]
		add	bx,2
		and	bx,0001Fh
		mov	[ds:bx+_traceBfr],si
		mov	[ds:_tracePtr],bx
  endif
 endif
		mov	bh,bh
		mov	bl,[fs:si]		; get opcode
		add	bx,bx			; index word table
		jmp	[cs:bx+opCodeTblBB]	; jump to opcode routine

; Exit back to Debugger

cineExit_BB:	mov	[ds:_eCState],STATE_BB
		jmp	cineExit

		endp

label		opCodeJmp	word

		include	"opcode.cod"		; opcode jump table

;----------------------------------------------------------------------------;
; All opcode routines follow, they are all lumped into a single 'proc'	     ;
; for ease of programming.						     ;
;----------------------------------------------------------------------------;
proc		opCodes		near

;----------------------------------------------------------------------------;
; *** INP *** (1x)							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opINP_A_AA	near
label		opINP_AA_AA	near
label		opINP_BB_AA	near

		mov	cl,[fs:si]		; get opcode

 ifdef LOG_IO_HITS
		SaveRegs			; save registers

		push	si			; pass PC
		push	cx			; pass OPCODE and PAGE
		push	0			; indicate I/O inputs
		call	_logIO			; log the I/O hit
		add	sp,6			; fix stack

		RestoreRegs			; restore registers
 endif

		inc	si			; inc PC
		and	cl,00Fh			; mask upper bits
		mov	bx,[ds:_ioInputs]	; get input values
		shr	bx,cl			; move to Bit-0
		and	bx,0001h		; remove other bits
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],ax		; save old Acc
		mov	ax,bx			; load into Acc
		jumpCineRet_AA			; return to loop

		SYNC4
label		opINP_B_AA	near

		mov	cl,[fs:si]		; get opcode

 ifdef LOG_IO_HITS
		SaveRegs			; save registers

		push	si			; pass PC
		push	cx			; pass OPCODE and PAGE
		push	1			; indicate I/O switches
		call	_logIO			; log the I/O hit
		add	sp,6			; fix stack

		RestoreRegs			; restore registers
 endif

		inc	si			; inc PC
		and	cl,007h			; mask upper bits
		mov	bl,[Byte ds:_ioSwitches] ; get switch values
		shr	bl,cl			; move to Bit-0
		and	bx,0001h		; remove other bits
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],dx		; save old Acc
		mov	dx,bx			; load into Acc
		jumpCineRet_AA			; return to loop

;----------------------------------------------------------------------------;
; *** OUT *** (9x)
;----------------------------------------------------------------------------;
		SYNC4

; Bi-Level brightness control

label		opOUTbi_A_A	near
label		opOUTbi_AA_A	near
label		opOUTbi_BB_A	near

		mov	cl,[fs:si]		; get opcode

 ifdef LOG_IO_HITS
		SaveRegs			; save registers

		push	si			; pass PC
		push	cx			; pass OPCODE and PAGE
		push	2			; indicate an output
		call	_logIO			; log the I/O hit
		add	sp,6			; fix stack

		RestoreRegs			; restore registers
 endif

		inc	si			; inc PC
		and	cl,007h			; mask unused bits
		cmp	cl,006h			; BRIGHT bit?
		jne	opOUTsnd_A		; no, do sound

		mov	cl,al			; get bit
		call	vgBrightBI		; call Bi-Level brightness
		jumpCineRet_A			; return to loop

; 16-Level brightness control

		SYNC4
label		opOUT16_A_A	near
label		opOUT16_AA_A	near
label		opOUT16_BB_A	near

		mov	cl,[fs:si]		; get opcode

 ifdef LOG_IO_HITS
		SaveRegs			; save registers

		push	si			; pass PC
		push	cx			; pass OPCODE and PAGE
		push	2			; indicate an output
		call	_logIO			; log the I/O hit
		add	sp,6			; fix stack

		RestoreRegs			; restore registers
 endif

		inc	si			; inc PC
		and	cl,007h			; mask unused bits
		cmp	cl,006h			; BRIGHT bit?
		jne	opOUTsnd_A		; no, do sound

		mov	cl,al			; set bit on/off
		call	vgBright16		; call 16-Level brightness
		jumpCineRet_A			; return to loop

; 64-Level brightness control

		SYNC4
label		opOUT64_A_A	near
label		opOUT64_AA_A	near
label		opOUT64_BB_A	near

		mov	cl,[fs:si]		; get opcode

 ifdef LOG_IO_HITS
		SaveRegs			; save registers

		push	si			; pass PC
		push	cx			; pass OPCODE and PAGE
		push	2			; indicate an output
		call	_logIO			; log the I/O hit
		add	sp,6			; fix stack

		RestoreRegs			; restore registers
 endif

		inc	si			; inc PC
		and	cl,007h			; mask unused bits
		cmp	cl,006h			; BRIGHT bit?
		jne	opOUTsnd_A		; no, do sound


		mov	cl,al			; set bit on/off
		call	vgBright64		; call 64-Level brightness
		jumpCineRet_A			; return to loop

; WotW Color

		SYNC4
label		opOUTWW_A_A	near
label		opOUTWW_AA_A	near
label		opOUTWW_BB_A	near

		mov	cl,[fs:si]		; get opcode

 ifdef LOG_IO_HITS
		SaveRegs			; save registers

		push	si			; pass PC
		push	cx			; pass OPCODE and PAGE
		push	2			; indicate an output
		call	_logIO			; log the I/O hit
		add	sp,6			; fix stack

		RestoreRegs			; restore registers
 endif

		inc	si			; inc PC
		and	cl,007h			; mask unused bits
		cmp	cl,006h			; BRIGHT bit?
		jne	opOUTsnd_A		; no, do sound

		mov	cl,al			; set bit on/off
		call	vgBrightCo		; call WotW color
		jumpCineRet_A			; return to loop

; Common routines...

opOUTsnd_A:	mov	bl,001h			; get bit mask
		shl	bl,cl			; shift into position
		test	al,001h			; set bitmask?
		jz	opOUT_Aset		; yes, set bit
		not	bl			; invert bit mask
		and	[ds:_ioOutputs],bl	; reset bit
		cmp	cl,005h			; Quarter off bit?
		je	opOUT_Aq		; Deal with quarter
		SaveRegs
		call	soundReset		; call sound board routines
		RestoreRegs
		jumpCineRet_A			; return to loop

		SYNC4
opOUT_Aset:	or	[ds:_ioOutputs],bl	; set bit mask
		SaveRegs
		call	soundSet		; call sound board routines
		RestoreRegs
		jumpCineRet_A			; return to loop

		SYNC4
opOUT_Aq:	or	[ds:_ioSwitches],SW_COIN ; reset coin counter
		jumpCineRet_A			; return to loop

; Bi-Level brigthness control

		SYNC4
label		opOUTbi_B_BB	near

		mov	cl,[fs:si]		; get opcode

 ifdef LOG_IO_HITS
		SaveRegs			; save registers

		push	si			; pass PC
		push	cx			; pass OPCODE and PAGE
		push	2			; indicate an output
		call	_logIO			; log the I/O hit
		add	sp,6			; fix stack

		RestoreRegs			; restore registers
 endif

		inc	si			; inc PC
		and	cl,007h			; mask unused bits
		cmp	cl,006h			; BRIGHT bit?
		jne	opOUTsnd_B		; no, do sound

		mov	cl,dl			; get brightness bit
		call	vgBrightBI		; call Bi-Level brightness
		jumpCineRet_BB			; return to loop

; 16-Level brigthness control

label		opOUT16_B_BB	near

		mov	cl,[fs:si]		; get opcode

 ifdef LOG_IO_HITS
		SaveRegs			; save registers

		push	si			; pass PC
		push	cx			; pass OPCODE and PAGE
		push	2			; indicate an output
		call	_logIO			; log the I/O hit
		add	sp,6			; fix stack

		RestoreRegs			; restore registers
 endif

		inc	si			; inc PC
		and	cl,007h			; mask unused bits
		cmp	cl,006h			; BRIGHT bit?
		jne	opOUTsnd_B		; no, do sound

		mov	cl,dl			; set bit on/off
		call	vgBright16		; call 16-Level brightness
		jumpCineRet_BB			; return to loop

; 64-Level brigthness control

label		opOUT64_B_BB	near

		mov	cl,[fs:si]		; get opcode

 ifdef LOG_IO_HITS
		SaveRegs			; save registers

		push	si			; pass PC
		push	cx			; pass OPCODE and PAGE
		push	2			; indicate an output
		call	_logIO			; log the I/O hit
		add	sp,6			; fix stack

		RestoreRegs			; restore registers
 endif

		inc	si			; inc PC
		and	cl,007h			; mask unused bits
		cmp	cl,006h			; BRIGHT bit?
		jne	opOUTsnd_B		; no, do sound

		mov	cl,dl			; set bit on/off
		call	vgBright64		; call 64-Level brightness
		jumpCineRet_BB			; return to loop

; WotW Color control

label		opOUTWW_B_BB	near

		mov	cl,[fs:si]		; get opcode

 ifdef LOG_IO_HITS
		SaveRegs			; save registers

		push	si			; pass PC
		push	cx			; pass OPCODE and PAGE
		push	2			; indicate an output
		call	_logIO			; log the I/O hit
		add	sp,6			; fix stack

		RestoreRegs			; restore registers
 endif

		inc	si			; inc PC
		and	cl,007h			; mask unused bits
		cmp	cl,006h			; BRIGHT bit?
		jne	opOUTsnd_B		; no, do sound

		mov	cl,dl			; set bit on/off
		call	vgBrightCo		; call WotW color
		jumpCineRet_BB			; return to loop

; Common routines...

opOUTsnd_B:	mov	bl,001h			; get bit mask
		shl	bl,cl			; shift into position
		test	dl,001h			; set bitmask?
		jz	opOUT_Bset		; yes, set bit
		not	bl			; invert bit mask
		and	[ds:_ioOutputs],bl	; reset bit
		cmp	cl,005h			; Quarter off bit?
		je	opOUT_Bq		; Deal with quarter
		SaveRegs
		call	soundSet		; call sound board routines
		RestoreRegs
		jumpCineRet_BB			; return to loop

		SYNC4
opOUT_Bset:	or	[ds:_ioOutputs],bl	; set bit mask
		SaveRegs
		call	soundReset		; call sound board routines
		RestoreRegs
		jumpCineRet_BB			; return to loop

		SYNC4
opOUT_Bq:	or	[ds:_ioSwitches],SW_COIN ; reset coin counter
		jumpCineRet_BB			; return to loop

;----------------------------------------------------------------------------;
; *** LDA imm *** (0x)							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opLDAimm_A_AA	near
label		opLDAimm_AA_AA	near
label		opLDAimm_BB_AA	near

		mov	bh,[fs:si]		; get imm value
		inc	si			; inc PC
		and	bx,0F00h		; mask
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],ax		; save old Acc
		mov	ax,bx			; load into Acc
		jumpCineRet_AA	

		SYNC4
label		opLDAimm_B_AA	near

		mov	bh,[fs:si]		; get imm value
		inc	si			; inc PC
		and	bx,0F00h		; mask
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],dx		; save old Acc
		mov	dx,bx			; load into Acc
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** LDA dir *** (Ax)							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opLDAdir_A_AA	near
label		opLDAdir_AA_AA	near
label		opLDAdir_BB_AA	near

		inc	si			; inc PC
		mov	bh,ch			; get P-reg
		shl	bl,3			; mask direct address
		shr	bx,3			; get word address
		lea	di,[ds:_ramBank+bx]	; save I-reg
		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],ax		; save old Acc
		mov	ax,bx			; load into Acc
		jumpCineRet_AA	

		SYNC4
label		opLDAdir_B_AA	near

		inc	si			; inc PC
		mov	bh,ch			; get P-reg
		shl	bl,3			; mask direct address
		shr	bx,3			; get word address
		lea	di,[ds:_ramBank+bx]	; save I-reg
		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],dx		; save old Acc
		mov	dx,bx			; load into Acc
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** LDA [i] ***							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opLDAirg_A_AA	near
label		opLDAirg_AA_AA	near
label		opLDAirg_BB_AA	near

		inc	si			; inc PC
		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],ax		; save old Acc
		mov	ax,bx			; load into Acc
		jumpCineRet_AA	

		SYNC4
label		opLDAirg_B_AA	near

		inc	si			; inc PC
		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],dx		; save old Acc
		mov	dx,bx			; load into Acc
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** ADD imm *** (2x)							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opADDimm_A_AA	near
label		opADDimm_AA_AA	near
label		opADDimm_BB_AA	near

		mov	bl,[fs:si]		; get imm value
		inc	si			; inc PC
		and	bx,000Fh		; mask
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	ax,bx			; add to Acc
		mov	[ds:_cFlag],ah		; save carry flag
		and	ah,00Fh			; mask carry
		jumpCineRet_AA	

		SYNC4
label		opADDimm_B_AA	near

		mov	bl,[fs:si]		; get imm value
		inc	si			; inc PC
		and	bx,000Fh		; mask
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		add	dx,bx			; add to Acc
		mov	[ds:_cFlag],dh		; save carry flag
		and	dh,00Fh			; mask carry
		jumpCineRet_AA	

		SYNC4
label		opADDimmX_A_AA	near
label		opADDimmX_AA_AA	near
label		opADDimmX_BB_AA	near

		mov	bl,[fs:si+1]		; get extended value
		add	si,2			; inc PC
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	ax,bx			; add to Acc
		mov	[ds:_cFlag],ah		; save carry flag
		and	ah,00Fh			; mask carry
		jumpCineRet_AA	

		SYNC4
label		opADDimmX_B_AA	near

		mov	bl,[fs:si+1]		; get extended value
		add	si,2			; inc PC
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		add	dx,bx			; add to Acc
		mov	[ds:_cFlag],dh		; save carry flag
		and	dh,00Fh			; mask carry
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** ADD dir ***							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opADDdir_A_AA	near
label		opADDdir_AA_AA	near
label		opADDdir_BB_AA	near

		inc	si			; inc PC
		mov	bh,ch			; get P-reg
		shl	bl,3			; mask direct address
		shr	bx,3			; get word address
		lea	di,[ds:_ramBank+bx]	; save I-reg
		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	ax,bx			; add to Acc
		mov	[ds:_cFlag],ah		; save carry flag
		and	ah,00Fh			; mask carry
		jumpCineRet_AA	

		SYNC4
label		opADDdir_B_AA	near

		inc	si			; inc PC
		mov	bh,ch			; get P-reg
		shl	bl,3			; mask direct address
		shr	bx,3			; get word address
		lea	di,[ds:_ramBank+bx]	; save I-reg
		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		add	dx,bx			; add to Acc
		mov	[ds:_cFlag],dh		; save carry flag
		and	dh,00Fh			; mask carry
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** ADD [i] ***							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opAWDirg_A_AA	near
label		opAWDirg_AA_AA	near
label		opAWDirg_BB_AA	near

label		opADDirg_A_AA	near
label		opADDirg_AA_AA	near
label		opADDirg_BB_AA	near

		inc	si			; inc PC
		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	ax,bx			; add to Acc
		mov	[ds:_cFlag],ah		; save carry flag
		and	ah,00Fh			; mask carry
		jumpCineRet_AA	

		SYNC4
label		opAWDirg_B_AA	near
label		opADDirg_B_AA	near

		inc	si			; inc PC
		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		add	dx,bx			; add to Acc
		mov	[ds:_cFlag],dh		; save carry flag
		and	dh,00Fh			; mask carry
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** SUB imm *** (3x)							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opSUBimm_A_AA	near
label		opSUBimm_AA_AA	near
label		opSUBimm_BB_AA	near

		mov	bl,[fs:si]		; get imm value
		inc	si			; inc PC
		and	bx,000Fh		; mask
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		xor	bx,0FFFh		; ones compliment
		stc				; bring in carry
		adc	ax,bx			; add to Acc
		mov	[ds:_cFlag],ah		; save carry flag
		and	ah,00Fh			; mask carry
		jumpCineRet_AA	

		SYNC4
label		opSUBimm_B_AA	near

		mov	bl,[fs:si]		; get imm value
		inc	si			; inc PC
		and	bx,000Fh		; mask
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		xor	bx,0FFFh		; ones compliment
		stc				; bring in carry
		adc	dx,bx			; add to Acc
		mov	[ds:_cFlag],dh		; save carry flag
		and	dh,00Fh			; mask carry
		jumpCineRet_AA	

		SYNC4
label		opSUBimmX_A_AA	near
label		opSUBimmX_AA_AA	near
label		opSUBimmX_BB_AA	near

		mov	bl,[fs:si+1]		; get imm value
		add	si,2			; inc PC
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		xor	bx,0FFFh		; ones compliment
		stc				; bring in carry
		adc	ax,bx			; add to Acc
		mov	[ds:_cFlag],ah		; save carry flag
		and	ah,00Fh			; mask carry
		jumpCineRet_AA	

		SYNC4
label		opSUBimmX_B_AA	near

		mov	bl,[fs:si+1]		; get imm value
		add	si,2			; inc PC
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		xor	bx,0FFFh		; ones compliment
		stc				; bring in carry
		adc	dx,bx			; add to Acc
		mov	[ds:_cFlag],dh		; save carry flag
		and	dh,00Fh			; mask carry
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** SUB dir ***							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opSUBdir_A_AA	near
label		opSUBdir_AA_AA	near
label		opSUBdir_BB_AA	near

		inc	si			; inc PC
		mov	bh,ch			; get P-reg
		shl	bl,3			; mask direct address
		shr	bx,3			; get word address
		lea	di,[ds:_ramBank+bx]	; save I-reg
		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		xor	bx,0FFFh		; ones compliment
		stc				; bring in carry
		adc	ax,bx			; add to Acc
		mov	[ds:_cFlag],ah		; save carry flag
		and	ah,00Fh			; mask carry
		jumpCineRet_AA	

		SYNC4
label		opSUBdir_B_AA	near

		inc	si			; inc PC
		mov	bh,ch			; get P-reg
		shl	bl,3			; mask direct address
		shr	bx,3			; get word address
		lea	di,[ds:_ramBank+bx]	; save I-reg
		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		xor	bx,0FFFh		; ones compliment
		stc				; bring in carry
		adc	dx,bx			; add to Acc
		mov	[ds:_cFlag],dh		; save carry flag
		and	dh,00Fh			; mask carry
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** SUB [i] ***							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opSUBirg_A_AA	near
label		opSUBirg_AA_AA	near
label		opSUBirg_BB_AA	near

		inc	si			; inc PC
		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		xor	bx,0FFFh		; ones compliment
		stc				; bring in carry
		adc	ax,bx			; add to Acc
		mov	[ds:_cFlag],ah		; save carry flag
		and	ah,00Fh			; mask carry
		jumpCineRet_AA	

		SYNC4
label		opSUBirg_B_AA	near

		inc	si			; inc PC
		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		xor	bx,0FFFh		; ones compliment
		stc				; bring in carry
		adc	dx,bx			; add to Acc
		mov	[ds:_cFlag],dh		; save carry flag
		and	dh,00Fh			; mask carry
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** CMP dir ***							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opCMPdir_A_AA	near
label		opCMPdir_AA_AA	near
label		opCMPdir_BB_AA	near

		inc	si			; inc PC
		mov	bh,ch			; get P-reg
		shl	bl,3			; mask direct address
		shr	bx,3			; get word address
		lea	di,[ds:_ramBank+bx]	; save I-reg
		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		xor	bx,0FFFh		; ones compliment
		stc				; bring in carry
		adc	bx,ax			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		jumpCineRet_AA	

		SYNC4
label		opCMPdir_B_AA	near

		inc	si			; inc PC
		mov	bh,ch			; get P-reg
		shl	bl,3			; mask direct address
		shr	bx,3			; get word address
		lea	di,[ds:_ramBank+bx]	; save I-reg
		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		xor	bx,0FFFh		; ones compliment
		stc				; bring in carry
		adc	bx,dx			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** AND [i] ***							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opANDirg_A_AA	near
label		opANDirg_AA_AA	near
label		opANDirg_BB_AA	near

		inc	si			; inc PC
		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],ax		; save old Acc
		and	ax,bx			; and with Acc
		jumpCineRet_AA	

		SYNC4
label		opANDirg_B_AA	near

		inc	si			; inc PC
		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],dx		; save old Acc
		and	dx,bx			; and with Acc
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** LDJ imm ***							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opLDJimm_A_A	near
label		opLDJimm_AA_A	near
label		opLDJimm_BB_A	near

		mov	bh,[fs:si+1]		; get high stuff
		add	si,2			; inc PC
		shl	bl,3			; mask lower nibble
		rol	bh,4			; swap nibbles
		shr	bx,4			; create address
		mov	bp,bx			; save in J-reg
		jumpCineRet_A	

		SYNC4
label		opLDJimm_B_BB	near

		mov	bh,[fs:si+1]		; get high stuff
		add	si,2			; inc PC
		shl	bl,3			; mask lower nibble
		rol	bh,4			; swap nibbles
		shr	bx,4			; create address
		mov	bp,bx			; save in J-reg
		jumpCineRet_BB	

;----------------------------------------------------------------------------;
; *** LDJ irg ***							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opLDJirg_A_A	near
label		opLDJirg_AA_A	near
label		opLDJirg_BB_A	near

		inc	si			; inc PC
		mov	bp,[ds:di]		; get jump address
		jumpCineRet_A	

		SYNC4
label		opLDJirg_B_BB	near

		inc	si			; inc PC
		mov	bp,[ds:di]		; get jump address
		jumpCineRet_BB	

;----------------------------------------------------------------------------;
; *** LDP imm ***							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opLDPimm_A_A	near
label		opLDPimm_AA_A	near
label		opLDPimm_BB_A	near

		mov	ch,[fs:si]		; get page value
		inc	si			; inc PC
		and	ch,00Fh			; mask high bits
		jumpCineRet_A	

		SYNC4
label		opLDPimm_B_BB	near

		mov	ch,[fs:si]		; get page value
		inc	si			; inc PC
		and	ch,00Fh			; mask high bits
		jumpCineRet_BB	

;----------------------------------------------------------------------------;
; *** LDI dir ***							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opLDIdir_A_A	near
label		opLDIdir_AA_A	near
label		opLDIdir_BB_A	near

		inc	si			; inc PC
		mov	bh,ch			; get P-reg
		shl	bl,3			; mask direct address
		shr	bx,3			; get word address
		mov	di,[ds:_ramBank+bx]	; get memory value
		and	di,000FFh		; can't access outside 256 regs
		add	di,di			; make word pointer
		add	di,Offset ds:_ramBank	; add RAM offset
		jumpCineRet_A	

		SYNC4
label		opLDIdir_B_BB	near

		inc	si			; inc PC
		mov	bh,ch			; get P-reg
		shl	bl,3			; mask direct address
		shr	bx,3			; get word address
		mov	di,[ds:_ramBank+bx]	; get memory value
		and	di,000FFh		; can't access outside 256 regs
		add	di,di			; make word pointer
		add	di,Offset _ramBank	; add RAM offset
		jumpCineRet_BB	

;----------------------------------------------------------------------------;
; *** STA dir ***							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opSTAdir_A_A	near
label		opSTAdir_AA_A	near
label		opSTAdir_BB_A	near

		inc	si			; inc PC
		mov	bh,ch			; get P-reg
		shl	bl,3			; mask direct address
		shr	bx,3			; get word address
		lea	di,[ds:_ramBank+bx]	; save I-reg

 ifdef DEBUGGER
  if WRITEADR
		cmp	di,[ds:_writeAdr]
		jne	opSTAd1
		mov	[ds:_insCount],1		; break on next check
opSTAd1:
  endif
 endif

		mov	[ds:di],ax		; save ACC
		jumpCineRet_A	

		SYNC4
label		opSTAdir_B_BB	near

		inc	si			; inc PC
		mov	bh,ch			; get P-reg
		shl	bl,3			; mask direct address
		shr	bx,3			; get word address
		lea	di,[ds:_ramBank+bx]	; save I-reg

 ifdef DEBUGGER
  if WRITEADR
		cmp	di,[ds:_writeAdr]
		jne	opSTAd2
		mov	[ds:_insCount],1		; break on next check
opSTAd2:
  endif
 endif

		mov	[ds:di],dx		; save ACC
		jumpCineRet_BB	

;----------------------------------------------------------------------------;
; *** STA irg ***							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opSTAirg_A_A	near
label		opSTAirg_AA_A	near
label		opSTAirg_BB_A	near

		inc	si			; inc PC

 ifdef DEBUGGER
  if WRITEADR
		cmp	di,[ds:_writeAdr]
		jne	opSTAi1
		mov	[ds:_insCount],1		; break on next check
opSTAi1:
  endif
 endif

		mov	[ds:di],ax		; save ACC
		jumpCineRet_A	

		SYNC4
label		opSTAirg_B_BB	near

		inc	si			; inc PC

 ifdef DEBUGGER
  if WRITEADR
		cmp	di,[ds:_writeAdr]
		jne	opSTAi2
		mov	[ds:_insCount],1		; break on next check
opSTAi2:
  endif
 endif

		mov	[ds:di],dx		; save ACC
		jumpCineRet_BB	

;----------------------------------------------------------------------------;
; *** XLT ***								     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opXLT_A_AA	near
label		opXLT_AA_AA	near
label		opXLT_BB_AA	near

		mov	bx,si			; get current address
		add	si,2			; inc PC
		and	bx,0F000h		; get just page
		or	bx,ax			; get address of ROM byte
		mov	bl,[fs:bx]		; get byte
		sub	bh,bh			; zero upper byte
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],ax		; save old Acc
		mov	ax,bx			; load into Acc
		jumpCineRet_AA	

		SYNC4
label		opXLT_B_AA	near

		mov	bx,si			; get current address
		add	si,2			; inc PC
		and	bx,0F000h		; get just page
		or	bx,dx			; get address of ROM byte
		mov	bl,[fs:bx]		; get byte
		sub	bh,bh			; zero upper byte
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],dx		; save old Acc
		mov	dx,bx			; load into Acc
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** MUL [i] ***							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opMULirg_A_AA	near
label		opMULirg_AA_AA	near
label		opMULirg_BB_AA	near

		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save value

		shl	bx,4			; shift into ADD position
		shl	dx,4			; get sign bit at bit 15
		or	dl,ah			; bring in top nibble of A
		mov	ah,al			; shift over 8 bits
		mov	al,[fs:si]		; get opcode

		inc	si
		cmp	[Byte fs:si],al		; another MUL follows?
		jne	opMUL1			; no, do only one

		inc	si
		cmp	[Byte fs:si],al		; another MUL follows?
		jne	opMUL2			; no, finish up

		inc	si
		cmp	[Byte fs:si],al		; another MUL follows?
		jne	opMUL3			; no, finish up

		inc	si
		cmp	[Byte fs:si],al		; another MUL follows?
		jne	opMUL4			; no, finish up

		inc	si
		cmp	[Byte fs:si],al		; another MUL follows?
		jne	opMUL5			; no, finish up

		inc	si
		cmp	[Byte fs:si],al		; another MUL follows?
		jne	opMUL6			; no, finish up

		inc	si
		cmp	[Byte fs:si],al		; another MUL follows?
		jne	opMUL7			; no, finish up

		inc	si
		cmp	[Byte fs:si],al		; another MUL follows?
		jne	opMUL8			; no, finish up

		inc	si
		cmp	[Byte fs:si],al		; another MUL follows?
		jne	opMUL9			; no, finish up

		inc	si
		cmp	[Byte fs:si],al		; another MUL follows?
		jne	opMUL10			; no, finish up

		inc	si
		cmp	[Byte fs:si],al		; another MUL follows?
		jne	opMUL11			; no, finish up

opMUL12:	inc	si
		sar	dx,1			; shift right D-reg
		rcr	ax,1
		test	al,080h			; one bit shifted out?
		jz	opMUL11			; no, skip
		add	dx,bx			; add memory value

opMUL11:	sar	dx,1			; shift right D-reg
		rcr	ax,1
		test	al,080h			; one bit shifted out?
		jz	opMUL10			; no, skip
		add	dx,bx			; add memory value

opMUL10:	sar	dx,1			; shift right D-reg
		rcr	ax,1
		test	al,080h			; one bit shifted out?
		jz	opMUL9			; no, skip
		add	dx,bx			; add memory value

opMUL9:		sar	dx,1			; shift right D-reg
		rcr	ax,1
		test	al,080h			; one bit shifted out?
		jz	opMUL8			; no, skip
		add	dx,bx			; add memory value

opMUL8:		sar	dx,1			; shift right D-reg
		rcr	ax,1
		test	al,080h			; one bit shifted out?
		jz	opMUL7			; no, skip
		add	dx,bx			; add memory value

opMUL7:		sar	dx,1			; shift right D-reg
		rcr	ax,1
		test	al,080h			; one bit shifted out?
		jz	opMUL6			; no, skip
		add	dx,bx			; add memory value

opMUL6:		sar	dx,1			; shift right D-reg
		rcr	ax,1
		test	al,080h			; one bit shifted out?
		jz	opMUL5			; no, skip
		add	dx,bx			; add memory value

opMUL5:		sar	dx,1			; shift right D-reg
		rcr	ax,1
		test	al,080h			; one bit shifted out?
		jz	opMUL4			; no, skip
		add	dx,bx			; add memory value

opMUL4:		sar	dx,1			; shift right D-reg
		rcr	ax,1
		test	al,080h			; one bit shifted out?
		jz	opMUL3			; no, skip
		add	dx,bx			; add memory value

opMUL3:		sar	dx,1			; shift right D-reg
		rcr	ax,1
		test	al,080h			; one bit shifted out?
		jz	opMUL2			; no, skip
		add	dx,bx			; add memory value

opMUL2:		sar	dx,1			; shift right D-reg
		rcr	ax,1
		test	al,080h			; one bit shifted out?
		jz	opMUL1			; no, skip
		add	dx,bx			; add memory value

; Save the proper ACC and set the carry flag correctly

opMUL1:		test	ah,001h			; ONE bit shifted out?
		jnz	opMULshf		; yes, add to B-reg
		mov	al,ah
		mov	ah,dl
		mov	bx,ax
		and	bh,00Fh
		mov	[ds:_pa0],bl
		mov	[ds:_accVal],bx
		add	bx,[ds:_cmpVal]
		mov	[ds:_cFlag],bh
		shr	ax,1
		and	ah,00Fh
		sar	dx,5
		and	dh,00Fh
		jumpCineRet_AA	

		SYNC4
opMULshf:	mov	al,ah
		mov	ah,dl
		mov	[ds:_pa0],al		; set PA0 flag
		shr	ax,1			; shift Acc
		and	ah,00Fh			; zero high nibble
		sar	dx,4			; shift right to correct
		mov	[ds:_accVal],dx		; save ACC
		and	[ds:_accVal],00Fh	; mask high nibble
		sar	dx,1			; shift right once more
		and	dh,00Fh			; mask high nibble
		add	dx,[ds:_cmpVal]		; add memory
		mov	[ds:_cFlag],dh		; save carry
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

		SYNC4
label		opMULirg_B_AA	near

		inc	si
		mov	bx,[ds:di]		; get memory value
		mov	[ds:_cmpVal],bx		; save value
		mov	[ds:_accVal],dx		; save old Acc
		mov	[ds:_pa0],al
		shl	dx,4
		sar	dx,5
		test	al,001h			; ONE shifted out?
		jnz	opMULirgB1		; yes, add
		add	bx,dx			; Add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		jumpCineRet_AA	

		SYNC4
opMULirgB1:	add	dx,bx			; Add to Acc
		mov	[ds:_cFlag],dh		; save carry flag
		and	dh,00Fh			; mask carry
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** LSRe *** (EB)							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opLSRe_A_AA	near
label		opLSRe_AA_AA	near
label		opLSRe_BB_AA	near

		mov	bx,0BEBh		; data in Data register
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jz	opLSRe_A0		; yes, do multiples
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	bx,ax			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		shr	ax,1			; logical shift right ACC
		jumpCineRet_AA	

		SYNC4
opLSRe_A0:	shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRe_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRe_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRe_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRe_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRe_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRe_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRe_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRe_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRe_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRe_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC

; Shift once with flags

opLSRe_A1:	mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	bx,ax			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		shr	ax,1			; logical shift right ACC
		jumpCineRet_AA

		SYNC4
label		opLSRe_B_AA	near

		inc	si			; inc PC
		cmp	[Word fs:si],0EB57h	; another one?
		jz	opLSRe_B0		; yes, do multiples
		mov	bx,0BEBh		; data in Data register
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		add	bx,dx			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		shl	dx,4			; get sign bit
		sar	dx,5			; shift arithmetic right
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

		SYNC4
opLSRe_B0:	shl	dx,4			; get sign bit
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0EB57h	; another one?
		jnz	opLSRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0EB57h	; another one?
		jnz	opLSRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0EB57h	; another one?
		jnz	opLSRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0EB57h	; another one?
		jnz	opLSRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0EB57h	; another one?
		jnz	opLSRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0EB57h	; another one?
		jnz	opLSRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0EB57h	; another one?
		jnz	opLSRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0EB57h	; another one?
		jnz	opLSRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0EB57h	; another one?
		jnz	opLSRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0EB57h	; another one?
		jnz	opLSRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
opLSRe_B1:	shr	dx,4			; fix register

; Shift once with flags

opLSRe_B2:	mov	bx,0BEBh		; data in Data register
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		add	bx,dx			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		shl	dx,4			; get sign bit
		sar	dx,5			; shift arithmetic right
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** LSRf *** (FB)							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opLSRf_A_AA	near
label		opLSRf_AA_AA	near
label		opLSRf_BB_AA	near

		mov	bx,0BFBh		; data in Data register
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jz	opLSRf_A0		; yes, do multiples
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	bx,ax			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		shr	ax,1			; logical shift right ACC
		jumpCineRet_AA	

		SYNC4
opLSRf_A0:	shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRf_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRf_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRf_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRf_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRf_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRf_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRf_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRf_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRf_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSRf_A1		; no, do last one
		shr	ax,1			; shift right
		inc	si			; inc PC

; Shift once with flags

opLSRf_A1:	mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	bx,ax			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		shr	ax,1			; logical shift right ACC
		jumpCineRet_AA	

		SYNC4
label		opLSRf_B_AA	near

		inc	si			; inc PC
		cmp	[Word fs:si],0FB57h	; another one?
		jz	opLSRf_B0		; yes, do multiples
		mov	bx,0BFBh		; data in Data register
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		add	bx,dx			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		shl	dx,4
		sar	dx,5			; shift arithmetic right
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

		SYNC4
opLSRf_B0:	shl	dx,4			; get sign bit
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FB57h	; another one?
		jnz	opLSRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FB57h	; another one?
		jnz	opLSRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FB57h	; another one?
		jnz	opLSRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FB57h	; another one?
		jnz	opLSRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FB57h	; another one?
		jnz	opLSRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FB57h	; another one?
		jnz	opLSRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FB57h	; another one?
		jnz	opLSRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FB57h	; another one?
		jnz	opLSRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FB57h	; another one?
		jnz	opLSRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FB57h	; another one?
		jnz	opLSRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
opLSRf_B1:	shr	dx,4			; fix register

; Shift once with flags

opLSRf_B2:	mov	bx,0BFBh		; data in Data register
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		add	bx,dx			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		shl	dx,4
		sar	dx,5			; shift arithmetic right
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** LSLe *** (EC)							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opLSLe_A_AA	near
label		opLSLe_AA_AA	near
label		opLSLe_BB_AA	near

		mov	bx,0CECh		; data in Data register
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jz	opLSLe_A0		; yes, do multiples
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	bx,ax			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		add	ax,ax			; logical shift left ACC
		and	ah,00Fh			; mask high nibble
		jumpCineRet_AA	

		SYNC4
opLSLe_A0:	add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLe_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLe_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLe_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLe_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLe_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLe_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLe_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLe_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLe_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLe_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
opLSLe_A1:	and	ah,00Fh			; fix register

; Shift once with flags

opLSLe_A2:	mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	bx,ax			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		add	ax,ax			; logical shift left ACC
		and	ah,00Fh			; mask high nibble
		jumpCineRet_AA	

		SYNC4
label		opLSLe_B_AA	near

		inc	si			; inc PC
		cmp	[Word fs:si],0EC57h	; another one?
		jz	opLSLe_B0		; yes, do multiples
		mov	bx,0CECh		; data in Data register
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		add	bx,dx			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		add	dx,dx			; logical shift left ACC
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

		SYNC4
opLSLe_B0:	add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0EC57h	; another one?
		jnz	opLSLe_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0EC57h	; another one?
		jnz	opLSLe_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0EC57h	; another one?
		jnz	opLSLe_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0EC57h	; another one?
		jnz	opLSLe_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0EC57h	; another one?
		jnz	opLSLe_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0EC57h	; another one?
		jnz	opLSLe_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0EC57h	; another one?
		jnz	opLSLe_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0EC57h	; another one?
		jnz	opLSLe_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0EC57h	; another one?
		jnz	opLSLe_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0EC57h	; another one?
		jnz	opLSLe_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
opLSLe_B1:	and	dh,00Fh			; fix register

; Shift once with flags

opLSLe_B2:	mov	bx,0CECh		; data in Data register
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		add	bx,dx			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		add	dx,dx			; logical shift left ACC
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** LSLf *** (FC)							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opLSLf_A_AA	near
label		opLSLf_AA_AA	near
label		opLSLf_BB_AA	near

		mov	bx,0CFCh		; data in Data register
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jz	opLSLf_A0		; yes, do multiples
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	bx,ax			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		add	ax,ax			; logical shift left ACC
		and	ah,00Fh			; mask high nibble
		jumpCineRet_AA	

		SYNC4
opLSLf_A0:	add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLf_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLf_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLf_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLf_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLf_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLf_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLf_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLf_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLf_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLf_A1		; no, do last one
		add	ax,ax			; shift left
		inc	si			; inc PC
opLSLf_A1:	and	ah,00Fh			; fix register

; Shift once with flags

opLSLf_A2:	mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	bx,ax			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		add	ax,ax			; logical shift left ACC
		and	ah,00Fh			; mask high nibble
		jumpCineRet_AA	

		SYNC4
label		opLSLf_B_AA	near

		inc	si			; inc PC
		cmp	[Word fs:si],0FC57h	; another one?
		jz	opLSLf_B0		; yes, do multiples
		mov	bx,0CFCh		; data in Data register
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		add	bx,dx			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		add	dx,dx			; logical shift left ACC
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

		SYNC4
opLSLf_B0:	add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0FC57h	; another one?
		jnz	opLSLf_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0FC57h	; another one?
		jnz	opLSLf_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0FC57h	; another one?
		jnz	opLSLf_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0FC57h	; another one?
		jnz	opLSLf_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0FC57h	; another one?
		jnz	opLSLf_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0FC57h	; another one?
		jnz	opLSLf_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0FC57h	; another one?
		jnz	opLSLf_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0FC57h	; another one?
		jnz	opLSLf_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0FC57h	; another one?
		jnz	opLSLf_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
		cmp	[Word fs:si],0FC57h	; another one?
		jnz	opLSLf_B1		; no, do last one
		add	dx,dx			; shift left
		add	si,2			; inc PC
opLSLf_B1:	and	dh,00Fh			; fix register

; Shift once with flags

opLSLf_B2:	mov	bx,0CFCh		; data in Data register
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		add	bx,dx			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		add	dx,dx			; logical shift left ACC
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** ASRe *** (ED)							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opASRe_A_AA	near
label		opASRe_AA_AA	near
label		opASRe_BB_AA	near

		mov	bx,0DEDh		; data in Data register
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jz	opASRe_A0		; yes, do multiples
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],ax		; save old Acc
		shl	ax,4			; get sign bit
		sar	ax,5			; arithmetic shift right ACC
		and	ah,00Fh			; mask high nibble
		jumpCineRet_AA	

		SYNC4
opASRe_A0:	shl	ax,4			; get carry flag
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRe_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRe_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRe_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRe_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRe_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRe_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRe_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRe_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRe_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRe_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
opASRe_A1:	shr	ax,4			; fix register

; Shift once with flags

opASRe_A2:	mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],ax		; save old Acc
		shl	ax,4			; get sign bit
		sar	ax,5			; arithmetic shift right ACC
		and	ah,00Fh			; mask high nibble
		jumpCineRet_AA	

		SYNC4
label		opASRe_B_AA	near

		inc	si			; inc PC
		cmp	[Word fs:si],0ED57h	; another one?
		jz	opASRe_B0		; yes, do multiples
		mov	bx,0DEDh		; data in Data register
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],dx		; save old Acc
		shl	dx,4			; get sign bit
		sar	dx,5			; arithmetic shift right ACC
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

		SYNC4
opASRe_B0:	shl	dx,4			; get sign bit
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0ED57h	; another one?
		jnz	opASRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0ED57h	; another one?
		jnz	opASRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0ED57h	; another one?
		jnz	opASRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0ED57h	; another one?
		jnz	opASRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0ED57h	; another one?
		jnz	opASRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0ED57h	; another one?
		jnz	opASRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0ED57h	; another one?
		jnz	opASRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0ED57h	; another one?
		jnz	opASRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0ED57h	; another one?
		jnz	opASRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0ED57h	; another one?
		jnz	opASRe_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
opASRe_B1:	shr	dx,4			; fix register

; Shift once with flags

opASRe_B2:	mov	bx,0DEDh		; data in Data register
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],dx		; save old Acc
		shl	dx,4			; get sign bit
		sar	dx,5			; arithmetic shift right ACC
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** ASRf *** (FD)							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opASRf_A_AA	near
label		opASRf_AA_AA	near
label		opASRf_BB_AA	near

		mov	bx,0DFDh		; data in Data register
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jz	opASRf_A0		; yes, do multiples
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],ax		; save old Acc
		shl	ax,4			; get sign bit
		sar	ax,5			; arithmetic shift right ACC
		and	ah,00Fh			; mask high nibble
		jumpCineRet_AA	

		SYNC4
opASRf_A0:	shl	ax,4			; get carry flag
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRf_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRf_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRf_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRf_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRf_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRf_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRf_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRf_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRf_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRf_A1		; no, do last one
		sar	ax,1			; shift right
		inc	si			; inc PC
opASRf_A1:	shr	ax,4			; fix register

; Shift once with flags

opASRf_A2:	mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],ax		; save old Acc
		shl	ax,4			; get sign bit
		sar	ax,5			; arithmetic shift right ACC
		and	ah,00Fh			; mask high nibble
		jumpCineRet_AA	

		SYNC4
label		opASRf_B_AA	near

		inc	si			; inc PC
		cmp	[Word fs:si],0FD57h	; another one?
		jz	opASRf_B0		; yes, do multiples
		mov	bx,0DFDh		; data in Data register
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],dx		; save old Acc
		shl	dx,4			; get sign bit
		sar	dx,5			; arithmetic shift right ACC
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

		SYNC4
opASRf_B0:	shl	dx,4			; get sign bit
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FD57h	; another one?
		jnz	opASRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FD57h	; another one?
		jnz	opASRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FD57h	; another one?
		jnz	opASRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FD57h	; another one?
		jnz	opASRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FD57h	; another one?
		jnz	opASRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FD57h	; another one?
		jnz	opASRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FD57h	; another one?
		jnz	opASRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FD57h	; another one?
		jnz	opASRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FD57h	; another one?
		jnz	opASRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
		cmp	[Word fs:si],0FD57h	; another one?
		jnz	opASRf_B1		; no, do last one
		sar	dx,1			; shift right
		add	si,2			; inc PC
opASRf_B1:	shr	dx,4			; fix register

; Shift once with flags

opASRf_B2:	mov	bx,0DFDh		; data in Data register
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0Cf],ax		; no carry, save PA0
		mov	[ds:_accVal],dx		; save old Acc
		shl	dx,4			; get sign bit
		sar	dx,5			; arithmetic shift right ACC
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** ASRDe *** (EE)							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opASRDe_A_AA	near
label		opASRDe_AA_AA	near
label		opASRDe_BB_AA	near

		mov	bx,0EEEh
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jz	opASRDe_A0		; yes, do multiples
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	bx,ax			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		shl	ax,4			; line up ACC with B-reg
		shl	dx,4			; get sign bit
		sar	dx,5			; arithmetic shift right
		rcr	ax,1			; shift into A-reg
		shr	ax,4			; fix A-reg
		and	dh,00Fh			; mask high nibble B-reg
		jumpCineRet_AA	

opASRDe_A0:	shl	ax,4			; get sign flags
		shl	dx,4
		sar	dx,5			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDe_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDe_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDe_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDe_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDe_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDe_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDe_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDe_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDe_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDe_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
opASRDe_A1:	shr	ax,4			; fix register

; Do last shift with flags

		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	bx,ax			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		shl	ax,4			; line up ACC with B-reg
		sar	dx,1			; arithmetic shift right
		rcr	ax,1			; shift into A-reg
		shr	ax,4			; fix A-reg
		and	dh,00Fh			; mask high nibble B-reg
		jumpCineRet_AA	

		SYNC4
label		opASRDe_B_AA	near

		inc	si			; inc PC
		mov	bx,0EEEh		; data in Data register
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		add	bx,dx			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		shl	dx,4			; get sign bit
		sar	dx,5			; arithmetic shift right
		and	dh,00Fh			; mask high nibble B-reg
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** ASRDf *** (FE)							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opASRDf_A_AA	near
label		opASRDf_AA_AA	near
label		opASRDf_BB_AA	near

		mov	bx,0EFEh
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jz	opASRDf_A0		; yes, do multiples
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	bx,ax			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		shl	ax,4			; line up ACC with B-reg
		shl	dx,4			; get sign bit
		sar	dx,5			; arithmetic shift right
		rcr	ax,1			; shift into A-reg
		shr	ax,4			; fix A-reg
		and	dh,00Fh			; mask high nibble B-reg
		jumpCineRet_AA	

opASRDf_A0:	shl	ax,4			; get sign flags
		shl	dx,4
		sar	dx,5			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDf_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDf_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDf_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDf_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDf_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDf_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDf_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDf_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDf_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opASRDf_A1		; no, do last shift
		sar	dx,1			; shift registers right
		rcr	ax,1
		inc	si			; inc PC
opASRDf_A1:	shr	ax,4			; fix register

; Do last shift with flags

		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	bx,ax			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		shl	ax,4			; line up ACC with B-reg
		sar	dx,1			; arithmetic shift right
		rcr	ax,1			; shift into A-reg
		shr	ax,4			; fix A-reg
		and	dh,00Fh			; mask high nibble B-reg
		jumpCineRet_AA	

		SYNC4
label		opASRDf_B_AA	near

		inc	si			; inc PC
		mov	bx,0EFEh
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		add	bx,dx			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		shl	dx,4			; get sign bit
		sar	dx,5			; arithmetic shift right
		and	dh,00Fh			; mask high nibble B-reg
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** LSLDe *** (EF)							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opLSLDe_A_AA	near
label		opLSLDe_AA_AA	near
label		opLSLDe_BB_AA	near

		mov	bx,0FEFh
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jz	opLSLDe_A0		; yes, do multiples
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	bx,ax			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		add	ax,ax			; logical shift left ACC
		and	ah,00Fh			; mask high nibble
		add	dx,dx			; logical shift left ACC
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

opLSLDe_A0:	add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDe_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDe_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDe_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDe_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDe_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDe_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDe_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDe_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDe_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDe_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
opLSLDe_A1:	and	ah,00Fh			; fix registers
		and	dh,00Fh

; Do last shift with flags
		
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	bx,ax			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		add	ax,ax			; logical shift left ACC
		and	ah,00Fh			; mask high nibble
		add	dx,dx			; logical shift left ACC
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

		SYNC4
label		opLSLDe_B_AA	near

		inc	si			; inc PC
		mov	bx,0FEFh
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		add	bx,dx			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		add	dx,dx			; logical shift left ACC
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** LSLDf *** (FF)							     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opLSLDf_A_AA	near
label		opLSLDf_AA_AA	near
label		opLSLDf_BB_AA	near

		mov	bx,0FFFh
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jz	opLSLDf_A0		; yes, do multiples
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	bx,ax			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		add	ax,ax			; logical shift left ACC
		and	ah,00Fh			; mask high nibble
		add	dx,dx			; logical shift left ACC
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

opLSLDf_A0:	add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDf_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDf_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDf_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDf_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDf_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDf_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDf_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDf_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDf_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
		cmp	[fs:si],bl		; another one?
		jnz	opLSLDf_A1		; no, do last one
		add	ax,ax			; shift registers left
		add	dx,dx
		inc	si			; inc PC
opLSLDf_A1:	and	ah,00Fh			; fix registers
		and	dh,00Fh

; Do last shift with flags
		
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],ax		; save old Acc
		add	bx,ax			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		add	ax,ax			; logical shift left ACC
		and	ah,00Fh			; mask high nibble
		add	dx,dx			; logical shift left ACC
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

		SYNC4
label		opLSLDf_B_AA	near

		inc	si			; inc PC
		mov	bx,0FFFh
		mov	[ds:_cmpVal],bx		; save Cmp value
		mov	[ds:_pa0],al		; save PA0
		mov	[ds:_accVal],dx		; save old Acc
		add	bx,dx			; add to Acc
		mov	[ds:_cFlag],bh		; save carry flag
		add	dx,dx			; logical shift left ACC
		and	dh,00Fh			; mask high nibble
		jumpCineRet_AA	

;----------------------------------------------------------------------------;
; *** JMPs ***								     ;
;----------------------------------------------------------------------------;
		SYNC4
label		opJMP_A_A	near
label		opJMP_AA_A	near
label		opJMP_BB_A	near

		and	si,0F000h		; remove old offset
		or	si,bp			; jump to new address
		jumpCineRet_A	

		SYNC4
label		opJMP_B_BB	near

		and	si,0F000h		; remove old offset
		or	si,bp			; jump to new address
		jumpCineRet_BB	

; Previous instruction was not an ACC instruction, nor was the instruction
; twice back a USB, therefore the minus flag test the current A-reg
 
		SYNC4
label		opJMI_A_A	near

		test	ah,008h			; A-reg negative?
		jnz	opJMP_A_A		; yes, jump
		inc	si			; inc PC
		jumpCineRet_A	

		SYNC4
label		opJEI_A_A	near
label		opJEI_AA_A	near
label		opJEI_BB_A	near

		les	bx,[ds:vgVecPtr]	; point to vector array
		mov	bx,[es:bx]		; get current X-DAC value
		shl	bx,4			; sign extend
		sar	bx,4

		test	[ds:_ioOutputs],080h	; test bit-7
		jz	opja1			; test Y-joystick

		cmp	bx,[ds:JoyX]		; for now, check sign
		jg	opJMP_A_A		; jump on not sign
		inc	si			; inc PC
		jumpCineRet_A

opja1:		cmp	bx,[ds:JoyY]		; for now, check sign
		jg	opJMP_A_A		; jump on not sign
		inc	si			; inc PC
		jumpCineRet_A

; Previous instruction WAS an ACC instruction, therefore the minus flag
; is that of the previous value of the effected register, save in 'accVal'.

		SYNC4
label		opJMI_AA_A	near
		test	[ds:_accVal+1],008h	; Previous ACC negative?
		jnz	opJMP_AA_A		; yes, jump
		inc	si			; inc PC
		jumpCineRet_A	


; Two instructions ago a ACC instruction using the B-reg, therefore the
; B-reg would contian the minus flag.

		SYNC4
label		opJMI_BB_A	near
		test	dh,008h			; B-reg negative?
		jnz	opJMP_BB_A		; yes, jump
		inc	si			; inc PC
		jumpCineRet_A	

; The last instruction was a 'USB', therefore the value of the minus
; flags must be in the A-reg.

		SYNC4
label		opJMI_B_BB	near
		test	ah,008h			; A-reg negative?
		jnz	opJMP_B_BB		; yes, jump
		inc	si			; inc PC
		jumpCineRet_BB	

label		opJEI_B_BB	near

		les	bx,[ds:vgVecPtr]	; point to vector array
		mov	bx,[es:bx]		; get current X-DAC value
		shl	bx,4			; sign extend
		sar	bx,4

		test	[ds:_ioOutputs],080h	; test bit-7
		jz	opjbb1			; test Y-joystick

		cmp	bx,[ds:JoyX]		; for now, check sign
		jg	opJMP_B_BB		; jump on not sign
		inc	si			; inc PC
		jumpCineRet_BB

opjbb1:		cmp	bx,[ds:JoyY]		; for now, check sign
		jg	opJMP_B_BB		; jump on not sign
		inc	si			; inc PC
		jumpCineRet_BB

		SYNC4
label		opJLT_A_A	near
label		opJLT_AA_A	near
label		opJLT_BB_A	near

		mov	bx,[ds:_cmpVal]		; get compare value
		cmp	bx,[ds:_accVal]		; is compare less than Acc?
		jb	opJMP_A_A		; yes, jump
		inc	si			; inc PC
		jumpCineRet_A	

		SYNC4
label		opJLT_B_BB	near

		mov	bx,[ds:_cmpVal]		; get compare value
		cmp	bx,[ds:_accVal]		; is compare less than Acc?
		jb	opJMP_B_BB		; yes, jump
		inc	si			; inc PC
		jumpCineRet_BB	

		SYNC4
label		opJEQ_A_A	near
label		opJEQ_AA_A	near
label		opJEQ_BB_A	near

		mov	bx,[ds:_cmpVal]		; get compare value
		cmp	bx,[ds:_accVal]		; is compare less than Acc?
		je	opJMP_A_A		; yes, jump
		inc	si			; inc PC
		jumpCineRet_A	

		SYNC4
label		opJEQ_B_BB	near

		mov	bx,[ds:_cmpVal]		; get compare value
		cmp	bx,[ds:_accVal]		; is compare less than Acc?
		je	opJMP_B_BB		; yes, jump
		inc	si			; inc PC
		jumpCineRet_BB	

		SYNC4
label		opJA0_A_A	near
label		opJA0_AA_A	near
label		opJA0_BB_A	near

		test	[ds:_pa0],001h		; was PA0 set?
		jnz	opJMP_A_A		; yes, jump
		inc	si			; inc PC
		jumpCineRet_A	

		SYNC4
label		opJA0_B_BB	near

		test	[ds:_pa0],001h		; was PA0 set?
		jnz	opJMP_B_BB		; yes, jump
		inc	si			; inc PC
		jumpCineRet_BB	

		SYNC4
label		opJNC_A_A	near
label		opJNC_AA_A	near
label		opJNC_BB_A	near

		test	[ds:_cFlag],0F0h	; any carry?
		jz	opJMP_A_A		; no carry, jmp
		inc	si			; inc PC
		jumpCineRet_A	

		SYNC4
label		opJNC_B_BB	near

		test	[ds:_cFlag],0F0h	; any carry?
		jz	opJMP_B_BB		; no carry, jmp
		inc	si			; inc PC
		jumpCineRet_BB	

		SYNC4
label		opJDR_A_A	near
label		opJDR_AA_A	near
label		opJDR_BB_A	near


		; Calculate number of cycles executed since
		; last 'VDR' instruction, add two and use as
		; cycle count, never branch

		inc	si			; inc PC
		jumpCineRet_A	

		SYNC4
label		opJDR_B_BB	near

		; Calculate number of cycles executed since
		; last 'VDR' instruction, add two and use as
		; cycle count, if branch should have been taken,
		; jump to 'A' tables, else jump to 'BB'.

		inc	si			; inc PC
		jumpCineRet_BB	

		SYNC4
label		opNOP_A_A	near
label		opNOP_AA_A	near
label		opNOP_BB_A	near

		inc	si			; inc PC
		jumpCineRet_A	

		SYNC4
label		opNOP_B_BB	near

		inc	si			; inc PC
		jumpCineRet_BB	

; *** JMP's with B-reg setup

		SYNC4
label		opJPP32_A_B	near
label		opJPP32_AA_B	near
label		opJPP32_BB_B	near

		; 00 = Offset 0000h
		; 01 = Offset 1000h
		; 02 = Offset 2000h
		; 03 = Offset 3000h
		; 04 = Offset 4000h
		; 05 = Offset 5000h
		; 06 = Offset 6000h
		; 07 = Offset 7000h

		mov	bh,ch			; get page register
		and	bh,007h			; get just page information
		shl	bh,4			; get proper offset
		mov	si,bx			; save page address

		; fall through

;		SYNC4
;label		opJMP_A_B	near

		and	si,0F000h		; remove old offset
		or	si,bp			; jump to new address
		jumpCineRet_B	

		SYNC4
label		opJPP32_B_BB	near

		; 00 = Offset 0000h
		; 01 = Offset 1000h
		; 02 = Offset 2000h
		; 03 = Offset 3000h
		; 04 = Offset 4000h
		; 05 = Offset 5000h
		; 06 = Offset 6000h
		; 07 = Offset 7000h

		mov	bh,ch			; get page register
		and	bh,007h			; get just page information
		shl	bh,4			; get proper offset
		mov	si,bx			; save page address
		and	si,0F000h		; remove old offset
		or	si,bp			; jump to new address
		jumpCineRet_BB	

		SYNC4
label		opJPP16_A_B	near
label		opJPP16_AA_B	near
label		opJPP16_BB_B	near

		; 00 = Offset 0000h
		; 01 = Offset 1000h
		; 02 = Offset 2000h
		; 03 = Offset 3000h

		mov	bh,ch			; get page register
		and	bh,003h			; get just page information
		shl	bh,4			; get proper offset
		mov	si,bx			; save page address

		; fall through

		SYNC4
label		opJMP_A_B	near

		and	si,0F000h		; remove old offset
		or	si,bp			; jump to new address
		jumpCineRet_B	

		SYNC4
label		opJPP16_B_BB	near

		; 00 = Offset 0000h
		; 01 = Offset 1000h
		; 02 = Offset 2000h
		; 03 = Offset 3000h

		mov	bh,ch			; get page register
		and	bh,003h			; get just page information
		shl	bh,4			; get proper offset
		mov	si,bx			; save page address
		and	si,0F000h		; remove old offset
		or	si,bp			; jump to new address
		jumpCineRet_BB	

		SYNC4
label		opJPP8_A_B	near
label		opJPP8_AA_B	near
label		opJPP8_BB_B	near

		mov	bl,ch			; get page register
		and	bl,003h			; get just page information
		jz	opJMP_A_B		; no change
		dec	bl			; page zero?
		jz	opJPP8_A1		; yes, use as is
		dec	bl			; page one?
		jz	opJPP8_AA		; yes, increment page
		xor	si,03000h		; flip pages

		; Fall through

;		SYNC4
;label		opJMP8_A_B	near

		and	si,0F000h		; remove old offset
		or	si,bp			; jump to new address
		jumpCineRet_B	

		SYNC4
opJPP8_A1:	mov	si,bp			; get new address with new page 0
		jumpCineRet_B	

		SYNC4
opJPP8_AA:	mov	si,bp			; get new address with no page
		add	si,01000h		; point to page one
		jumpCineRet_B	

		SYNC4
label		opJPP8_B_BB	near

		mov	bl,ch			; get page register
		and	bl,003h			; get just page information
		jz	opJMP_B_BB		; no change
		dec	bl			; page zero?
		jz	opJPP8_B1		; yes, use as is
		dec	bl			; page one?
		jz	opJPP8_BB		; yes, increment page
		xor	si,03000h		; flip pages

;		SYNC4
;label		opJMP8_B_BB	near

		and	si,0F000h		; remove old offset
		or	si,bp			; jump to new address
		jumpCineRet_BB	

		SYNC4
opJPP8_B1:	mov	si,bp			; get new address with new page
		jumpCineRet_BB	

		SYNC4
opJPP8_BB:	mov	si,bp			; get new address with no page
		add	si,01000h		; point to page one
		jumpCineRet_BB	

		SYNC4
label		opJMI_A_B	near
		test	ah,008h			; A-reg negative?
		jnz	opJMP_A_B
		inc	si			; inc PC
		jumpCineRet_B	

		SYNC4
label		opJEI_A_B	near
label		opJEI_AA_B	near
label		opJEI_BB_B	near

		les	bx,[ds:vgVecPtr]	; point to vector array
		mov	bx,[es:bx]		; get current X-DAC value
		sar	bx,4			; fix shift
	
		test	[ds:_ioOutputs],080h	; test bit-7
		jz	opjb1			; test Y-joystick

		cmp	bx,[ds:JoyX]		; for now, check sign
		jg	opJMP_A_B		; jump on not sign
		inc	si			; inc PC
		jumpCineRet_B

opjb1:		cmp	bx,[ds:JoyY]		; for now, check sign
		jg	opJMP_A_B		; jump on not sign
		inc	si			; inc PC
		jumpCineRet_B

		SYNC4
label		opJMI_AA_B	near
		test	[ds:_accVal+1],008h	; Previous ACC negative?
		jnz	opJMP_A_B
		inc	si			; inc PC
		jumpCineRet_B	

		SYNC4
label		opJMI_BB_B	near
		test	dh,008h			; B-reg negative?
		jnz	opJMP_A_B
		inc	si			; inc PC
		jumpCineRet_B	

		SYNC4
label		opJLT_A_B	near
label		opJLT_AA_B	near
label		opJLT_BB_B	near

		mov	bx,[ds:_cmpVal]		; get compare value
		cmp	bx,[ds:_accVal]		; is compare less than Acc?
		jb	opJMP_A_B		; yes, jump
		inc	si			; inc PC
		jumpCineRet_B	

		SYNC4
label		opJEQ_A_B	near
label		opJEQ_AA_B	near
label		opJEQ_BB_B	near

		mov	bx,[ds:_cmpVal]		; get compare value
		cmp	bx,[ds:_accVal]		; is compare less than Acc?
		je	opJMP_A_B		; yes, jump
		inc	si			; inc PC
		jumpCineRet_B	

		SYNC4
label		opJA0_A_B	near
label		opJA0_AA_B	near
label		opJA0_BB_B	near

		test	[ds:_pa0],001h		; was PA0 set?
		jnz	opJMP_A_B		; yes, jump
		inc	si			; inc PC
		jumpCineRet_B	

		SYNC4
label		opJNC_A_B	near
label		opJNC_AA_B	near
label		opJNC_BB_B	near

		test	[ds:_cFlag],0F0h	; any carry?
		jz	opJMP_A_B		; no Carry, jmp
		inc	si			; inc PC
		jumpCineRet_B	

		SYNC4
label		opJDR_A_B	near
label		opJDR_AA_B	near
label		opJDR_BB_B	near

		; Calculate number of cycles executed since
		; last 'VDR' instruction, add two and use as
		; cycle count, never branch

		jumpCineRet_B	

		SYNC4
label		opNOP_A_B	near
label		opNOP_AA_B	near
label		opNOP_BB_B	near

		inc	si			; inc PC
		jumpCineRet_B	

; Video Routines.

		SYNC4
label		opLLT_A_AA	near
label		opLLT_AA_AA	near
label		opLLT_BB_AA	near

		inc	si		; inc PC
		sub	bx,bx		; zero index
opLLTa1:	mov	bh,ah		; get PA11 and PA9
		and	bh,00Ah		; mask all other bits
		jz	opLLTa2		; if zero, then no mismatch
		xor	bh,00Ah		; flip bits
		jnz	opLLTa4		; if not zero, then mismatch found
opLLTa2:	mov	bh,dh		; get SA11 and SA9
		and	bh,00Ah		; mask all other bits
		jz	opLLTa3		; if zero, then no mismatch
		xor	bh,00Ah		; flip bits
		jnz	opLLTa4		; if not zero, then mismatch found

opLLTa3:	add	ax,ax		; shift A-reg
		add	dx,dx		; shift B-reg
		inc	bl		; add to index
		jnz	opLLTa1		; try again
		jumpCineRet_AA		; if overflow, just return

		SYNC4
opLLTa4:	and	ah,00Fh		; mask upper bits of A-reg
		and	dh,00Fh		; mask upper bits of B-reg
		SaveRegs
		call	vgLength	; call with index in BL-reg
		RestoreRegs
opLLTaErr:	jumpCineRet_AA	

		SYNC4
label		opLLT_B_AA	near

		inc	si		; inc PC
		sub	bx,bx		; zero index
opLLTb1:	mov	bh,ah		; get PA11 and PA9
		and	bh,00Ah		; mask all other bits
		jz	opLLTb2		; if zero, then no mismatch
		xor	bh,00Ah		; flip bits
		jnz	opLLTb4		; if not zero, then mismatch found
opLLTb2:	mov	bh,dh		; get SA11 and SA9
		and	bh,00Ah		; mask all other bits
		jz	opLLTb3		; if zero, then no mismatch
		xor	bh,00Ah		; flip bits
		jnz	opLLTb4		; if not zero, then mismatch found

opLLTb3:	add	dx,dx		; shift B-reg
		inc	bl		; add to index
		jnz	opLLTb2		; try again (just B-reg)
		jumpCineRet_AA		; if overflow, just return

		SYNC4
opLLTb4:	and	dh,00Fh		; mask upper bits of B-reg
		SaveRegs
		call	vgLength	; call with index in BL-reg
		RestoreRegs
		jumpCineRet_AA	

		SYNC4
label		opVIN_A_A	near
label		opVIN_AA_A	near
label		opVIN_BB_A	near
		inc	si			; inc PC
		SaveRegs
		call	vgStart
		RestoreRegs
		jumpCineRet_A	

		SYNC4
label		opVIN_B_BB	near
		inc	si			; inc PC
		SaveRegs
		call	vgStart
		RestoreRegs
		jumpCineRet_BB	

		SYNC4
label		opWAI_A_A	near
label		opWAI_AA_A	near
label		opWAI_BB_A	near

		push	ax
		push	cx
		push	dx
		push	bp
		push	di
		push	ds
		push	fs

; Check for RESET switch

		test	[ds:_ioSwitches],SW_RESET
		jnz	opWAI_Aexit

; Wait for refresh time

		mov	bl,[fs:si]		; get opcode
opWAI_Await:	inc	si			; point to next opcode
		mov	dx,[ds:refreshRate]	; get clock increment
		test	dx,dx			; 0=No wait
		jz	opWAI_Anw		; no refresh wait

		cmp	bl,[fs:si]		; same?
		jne	opWAI_Awc1		; yes, use as is
		inc	si			; else skip 2nd WAI
opWAI_Awc1:	sub	al,al
		cli
		out	i_8254+3,al		; latch timer 0
		IOWAIT				; wait for I/O
		in	al,i_8254		; get timer
		IOWAIT				; wait for I/O
		xchg	al,ah
		in	al,i_8254
		sti

		xchg	al,ah
		neg	ax			; make it an up counter
		sub	ax,[ds:refreshCnt]	; get difference
		cmp	ax,dx
		jc	opWAI_Awc1		; wait...

		add	[ds:refreshCnt],dx	; update counter with inc
		
opWAI_Anw:	cmp	bl,[fs:si]		; another WAI inst?
		je	opWAI_Await		; yes, wait again

		push	si
		call	vgSwap
		call	soundIO
		pop	si
		pop	fs
		pop	ds
		pop	di
		pop	bp
		pop	dx
		pop	cx
		pop	ax
		jumpCineRet_A	

opWAI_Aexit:	inc	si		; skip the WAI instruction
		pop	fs
		pop	ds
		pop	di
		pop	bp
		pop	dx
		pop	cx
		pop	ax
		jmp	cineExit_A

		SYNC4
label		opWAI_B_BB	near

		push	ax
		push	cx
		push	dx
		push	bp
		push	di
		push	ds
		push	fs

; Check for RESET switch

		test	[ds:_ioSwitches],SW_RESET
		jnz	opWAI_Bexit

; Wait for refresh time

		mov	bl,[fs:si]		; get opcode
opWAI_Bwait:	inc	si			; point to next opcode
		mov	dx,[ds:refreshRate]	; get clock increment
		test	dx,dx			; 0=No wait
		jz	opWAI_Bnw		; no refresh wait

		cmp	bl,[fs:si]		; same?
		jne	opWAI_Bwc1		; yes, use as is
		inc	si			; else skip 2nd WAI
opWAI_Bwc1:	sub	al,al
		cli
		out	i_8254+3,al		; latch timer 0
		IOWAIT				; wait for I/O
		in	al,i_8254		; get timer
		IOWAIT				; wait for I/O
		xchg	al,ah
		in	al,i_8254
		sti

		xchg	al,ah
		neg	ax			; make it an up counter
		sub	ax,[ds:refreshCnt]	; get difference
		cmp	ax,dx
		jc	opWAI_Bwc1		; wait...

		add	[ds:refreshCnt],dx	; update counter with inc
		
opWAI_Bnw:	cmp	bl,[fs:si]		; another WAI inst?
		je	opWAI_Bwait		; yes, wait again

		push	si
		call	vgSwap
		call	soundIO
		pop	si
		pop	fs
		pop	ds
		pop	di
		pop	bp
		pop	dx
		pop	cx
		pop	ax
		jumpCineRet_BB	

opWAI_Bexit:	inc	si		; skip the WAI instruction
		pop	fs
		pop	ds
		pop	di
		pop	bp
		pop	dx
		pop	cx
		pop	ax
		jmp	cineExit_BB

		SYNC4
label		opVDR_A_A	near
label		opVDR_AA_A	near
label		opVDR_BB_A	near

		inc	si			; inc PC
		SaveRegs
		call	vgDraw
		RestoreRegs
		jumpCineRet_A	

		SYNC4
label		opVDR_B_BB	near

		inc	si			; inc PC
		SaveRegs
		call	vgDraw
		RestoreRegs
		jumpCineRet_BB	

		endp

		ends
		end
