;----------------------------------------------------------------------------;
; Line routines for a VGA planar 4 bit (16 color) display using 64k banks.
;
; Author: Zonn Moore
; Date:   3/16/00
;
; (c) Copyright 2000, Zonn Moore.  All rights reserved.
;----------------------------------------------------------------------------;
		ideal
		P386N
		include	"zvga.inc"
		include	"video.inc"

macro		SVGAMEM	reg,val
		and	[reg],val
		endm

macro		SETBANK
		local	sb1
		cmp	dx,[ds:vCurBank]
		je	sb1
		call	vSetBank
sb1:
		endm

segment		_DATA	public 'DATA' dword
		extrn	_vgSeg:Word,vgVecPtr:Dword

		ends

segment		_TEXT	public 'CODE' dword
		assume	cs:_TEXT

		align	4
public		vSetupB4
proc		vSetupB4	near
		mov	dx,VGAGCTL
		SVGREG	vgrMODE,00Bh	; set write mode 3, read mode 1
		SVGREG	vgrDCARE,0	; don't care for any bits (read FF)
		SVGREG	vgrBM,0FFh	; all bits available
		SVGREG	vgrDR,000h	; ingore, no rotate
		ret
		endp

		align	4
public		vSetColorB4
proc		vSetColorB4	near
		mov	dx,VGAGCTL
		mov	ah,al
		SVGRAH	vgrSR,ah	; set color in Set/Reset
		ret
		endp

		align	4
public		vSetDrawB4
proc		vSetDrawB4	near
		ret			; do nothing for now
		endp

		align	4
public		vSetEraseB4
proc		vSetEraseB4	near
		mov	dx,VGAGCTL
		sub	ah,ah
		SVGRAH	vgrSR,ah	; set black in Set/Reset
		ret
		endp

;----------------------------------------------------------------------------;
; vXltBiB4								     ;
;									     ;
; Set brightness value based on a Bi-Level monitor.			     ;
;									     ;
; Called with CL - Bit0 with value of Brightness Bit			     ;
;----------------------------------------------------------------------------;
		align	4
public		vXltBiB4
proc		vXltBiB4	near
		mov	bl,007h		; get dim color
		and	cl,001h		; mask all other bits
		shl	cl,3		; move to bit 3
		or	bl,cl		; make bright if set (reset Z-flg)
		ret
		endp

;----------------------------------------------------------------------------;
; vgXlt16B4								     ;
;									     ;
; Set brightness value based on a 16-Level monitor.			     ;
; Only set color if bit-0 (Latch bit) of CL-reg is set.			     ;
;									     ;
; Brightness level in BX-register is:					     ;
;									     ;
;    xxxx xxxx xxxx CCCC						     ;
;									     ;
; Where:								     ;
;    CCCC = Intensity level.						     ;
;           0000 = Lowest intensity					     ;
;           0001 = 2nd lowest intensity					     ;
;           1111 = Highest intensity					     ;
;----------------------------------------------------------------------------;
		align	4
public		vXlt16B4
proc		vXlt16B4	near
		test	cl,1		; set Z-flg if bit reset
		jz	vg16L1		; no latch bit, skip
		and	bl,00Fh		; get color
		cmp	bl,1		; black not allowed
		adc	bl,0		; (resets Z-flg)
vg16L1:		ret
		endp

;----------------------------------------------------------------------------;
; vgBright64								     ;
;									     ;
; Set brightness value based on a 64-Level monitor.			     ;
; Only set color if bit-0 (Latch bit) of CL-reg is set.			     ;
;									     ;
; Brightness level in BX-register is:					     ;
;									     ;
;    xxxx xxxx CCCC CCxx						     ;
;									     ;
; Where:								     ;
;    CCCC CC = Intensity level.						     ;
;              1111 11 = Lowest intensity				     ;
;              1111 10 = 2nd lowest intensity				     ;
;              0000 00 = Highest intensity				     ;
;----------------------------------------------------------------------------;
		align	4
public		vXlt64B4
proc		vXlt64B4	near
		test	cl,1		; set Z-flg if bit reset
		jz	vg64L1		; no latch bit, skip
		not	bl		; inverted logic
		shr	bl,4		; get color (only 16 bits allowed)
		cmp	bl,1		; black not allowed
		adc	bl,0		; (resets Z-flg)
vg64L1:		ret
		endp

;----------------------------------------------------------------------------;
; vgBrightCo								     ;
;									     ;
; Set brightness value based on WotW color.				     ;
;									     ;
; Color in BX-register is:						     ;
;									     ;
;    xxxx BBBB GGGG RRRR						     ;
;									     ;
; Where:								     ;
;    BBBB = Blue intensity level.					     ;
;    GGGG = Green intensity level.					     ;
;    RRRR = Red intensity level.					     ;
;									     ;
;    1111 1111 1111 = Lowest intensity white.				     ;
;    1110 1111 1111 = Lowest intensity blue.				     ;
;    1111 1110 1111 = Lowest intensity green.				     ;
;    1111 1111 1110 = Lowest intensity red.				     ;
;    0000 0000 0000 = Brightest white.					     ;
;----------------------------------------------------------------------------;
		align	4
public		vXltColorB4
proc		vXltColorB4	Near

		test	cl,1			; set new color?
		jz	vgWWL1			; no, skip
		not	bx			; invert
		and	bx,00FFFh		; get only colors
		jz	vgcobl			; black
		test	bx,0888h		; any bright bits?
		jnz	vgcobr			; yes, do bright
		test	bx,0444h		; any dim bits?
		jz	vgcodm			; no, dim white

		shl	bx,6
		rcl	cl,1			; get blue 
		shl	bx,4
		rcl	cl,1			; get green
		shl	bx,4
		rcl	cl,1			; get red
		and	cl,007h
		mov	bl,cl			; get color
		ret

vgcobr:		shl	bx,5
		rcl	cl,1
		shl	bx,4
		rcl	cl,1
		shl	bx,4
		rcl	cl,1
		and	cl,007h
		add	cl,7
		mov	bl,cl			; get color
		ret

vgcobl:		mov	cl,0			; black
		ret

vgcodm:		mov	cl,00Fh			; dim white
vgWWL1:         ret
		endp

;----------------------------------------------------------------------------;
; Planar Clear a Window using 64k banks.				     ;
;									     ;
; Called with:								     ;
;    DS = Pointer Data Segment						     ;
;									     ;
; Assumes all windows start and stop on a 4 byte boundry.		     ;
;----------------------------------------------------------------------------;
		align	4
public		v64ClsB4
proc		v64ClsB4

		mov	ax,[ds:vCurSRow]	; get starting row of screen
		mul	[ds:vYBytes]		; dx:ax = index into video memory
		mov	di,ax			; dx:di = start of bank pointer
		SETBANK				; set to the proper starting page

		mov	es,[ds:vSeg]		; get video memory segment

		mov	dx,[Word ds:vWinSize]	; get size in BX:DX
		mov	bx,[Word ds:vWinSize+2]

		mov	eax,0FFFFFFFFh		; get mask

vcls1:		mov	si,di			; get offset
		neg	si			; get bytes to end of bank
		jnz	vcls2			; if bank not full, check size
		test	bx,bx			; .gt. 64k block?
		jz	vcls3			; no, write to end of window

		mov	cx,04000h		; dwords in a bank
	rep	stos	[Dword es:di]		; clear bank
		dec	bx			; fix count
		mov	si,dx
		or	si,bx			
		jz	vcls5			; if done, exit
		call	vIncBank		; point to next bank
		jmp	Short vcls1		; do next bank

vcls2:		test	bx,bx			; greater than 64k?
		jnz	vcls4			; yes, write to end of bank
		cmp	si,dx			; can we write to end of bank?
		jb	vcls4			; no, use count to end of bank
vcls3:		mov	si,dx			; else use count to end of window

vcls4:		mov	cx,si			; get bytes to end of bank
		shr	cx,2			; make dword counter
	rep	stos	[Dword es:di]		; clear to end of bank
		sub	dx,si			; subtract from count
		sbb	bx,0
		mov	si,dx
		or	si,bx			
		jz	vcls5			; if done, exit
		call	vIncBank
		jmp	Short vcls1		; keep going

vcls5:		ret

		endp

;----------------------------------------------------------------------------;
; v64LineB4								     ;
;									     ;
; Line draw routine using 64k banks.					     ;
;									     ;
; Called with								     ;
;    BX = xStart							     ;
;    CX = yStart							     ;
;    SI = xEnd								     ;
;    DI = yEnd								     ;
;									     ;
; Assumed:								     ;
;    DS = Segment of video memory.					     ;
;    GS = Data Segment.							     ;
;    Line has been properly clipped.					     ;
;    Write mode #3.							     ;
;    Color set in S/R register.						     ;
;									     ;
; Uses:									     ;
;    AX, BX, CX, DX							     ;
;----------------------------------------------------------------------------;
		align	4
public		v64LineB4
proc		v64LineB4	near

		cmp	di,cx			; horizontal line?
		je	v64HzLineB4
		cmp	si,bx			; vertical line?
		je	v64VtLineB4
		jmp	v64DgLineB4		; draw diagonal line
		endp

;----------------------------------------------------------------------------;
; v64VtLineB4								     ;
;									     ;
; Vertical line draw routine using 64k banks.				     ;
;									     ;
; Called with								     ;
;    BX = xStart							     ;
;    CX = yStart							     ;
;    SI = xEnd								     ;
;    DI = yEnd								     ;
;									     ;
; Assumed:								     ;
;    DS = Data Segment.							     ;
;    Line has been properly clipped.					     ;
;    Write mode #3.							     ;
;    Color set in S/R register.						     ;
;									     ;
; Uses:									     ;
;    AX, BX, CX, DX							     ;
;----------------------------------------------------------------------------;
		align	4
public		v64VtLineB4
proc		v64VtLineB4	near

; Get memory addresses

		mov	si,cx		; get yStart in SI-reg
		cmp	di,si		; get number of pixels to display
		jae	vpl0		; if no error, use as is
		xchg	si,di		; reverse end points
vpl0:		sub	di,si		; calc distance between end points
		inc	di		; get number of pixels

		add	si,[ds:vCurSRow]

		mov	ax,[ds:vYBytes]
		mul	si		; point to yStart
		mov	cl,bl		; get bit shift count
		shr	bx,3		; fix X value
		add	bx,ax		; get starting address in BX-reg
		adc	dx,0		; add overflow
		SETBANK			; set to the proper starting page

		and	cl,007h
		mov	dl,080h		; set bit mask
		shr	dl,cl		; shift to proper position

		mov	ax,[ds:vYBytes]	; get index

		push	ds
		mov	ds,[ds:vSeg]	; point to video memory

		mov	cx,di		; save count
		shr	di,3		; unwind loop to counts of 8
		jz	vpl2		; if no counts of eight, skip

vpl1:		SVGAMEM	bx,dl		; set pixel
		add	bx,ax		; add increment
		jc	vpl1Bank1	; adjust bank if needed
vpl1Ret1:	SVGAMEM	bx,dl
		add	bx,ax
		jc	vpl1Bank2
vpl1Ret2:	SVGAMEM	bx,dl
		add	bx,ax
		jc	vpl1Bank3
vpl1Ret3:	SVGAMEM	bx,dl
		add	bx,ax
		jc	vpl1Bank4

vpl1Ret4:	SVGAMEM	bx,dl
		add	bx,ax
		jc	vpl1Bank5
vpl1Ret5:	SVGAMEM	bx,dl
		add	bx,ax
		jc	vpl1Bank6
vpl1Ret6:	SVGAMEM	bx,dl
		add	bx,ax
		jc	vpl1Bank7
vpl1Ret7:	SVGAMEM	bx,dl
		add	bx,ax
		jc	vpl1Bank8

vpl1Ret8:	dec	di		; done?
		jnz	vpl1		; no, loop

vpl2:		and	cx,007h		; get remainder of pixels
		jz	vpl4		; if none, skip

		SVGAMEM	bx,dl
		add	bx,ax
		jc	vpl2Bank1
vpl2Ret1:	dec	cx		
		jz	vpl4		

		SVGAMEM	bx,dl
		add	bx,ax
		jc	vpl2Bank2
vpl2Ret2:	dec	cx		
		jz	vpl4		

		SVGAMEM	bx,dl
		add	bx,ax
		jc	vpl2Bank3
vpl2Ret3:	dec	cx		
		jz	vpl4		

		SVGAMEM	bx,dl
		add	bx,ax
		jc	vpl2Bank4
vpl2Ret4:	dec	cx		
		jz	vpl4		

		SVGAMEM	bx,dl
		add	bx,ax
		jc	vpl2Bank5
vpl2Ret5:	dec	cx		
		jz	vpl4		

		SVGAMEM	bx,dl
		add	bx,ax
		jc	vpl2Bank6
vpl2Ret6:	dec	cx		
		jz	vpl4		

		SVGAMEM	bx,dl
vpl4:		pop	ds
		ret

vpl1Bank1:	call	vIncBank
		jmp	vpl1Ret1

vpl1Bank2:	call	vIncBank
		jmp	vpl1Ret2

vpl1Bank3:	call	vIncBank
		jmp	vpl1Ret3

vpl1Bank4:	call	vIncBank
		jmp	vpl1Ret4

vpl1Bank5:	call	vIncBank
		jmp	vpl1Ret5

vpl1Bank6:	call	vIncBank
		jmp	vpl1Ret6

vpl1Bank7:	call	vIncBank
		jmp	vpl1Ret7

vpl1Bank8:	call	vIncBank
		jmp	vpl1Ret8


vpl2Bank1:	call	vIncBank
		jmp	vpl2Ret1

vpl2Bank2:	call	vIncBank
		jmp	vpl2Ret2

vpl2Bank3:	call	vIncBank
		jmp	vpl2Ret3

vpl2Bank4:	call	vIncBank
		jmp	vpl2Ret4

vpl2Bank5:	call	vIncBank
		jmp	vpl2Ret5

vpl2Bank6:	call	vIncBank
		jmp	vpl2Ret6

		endp

;----------------------------------------------------------------------------;
; v64PointB4								     ;
;									     ;
; Planar point draw routine, using 64k banks.				     ;
;									     ;
; Called with								     ;
;    BX = xStart							     ;
;    CX = yStart							     ;
;									     ;
; Assumed:								     ;
;    DS = Data Segment.							     ;
;    Write mode #3.							     ;
;    Color set in S/R register.						     ;
;									     ;
; Uses:									     ;
;    AX, BX, CX, DX							     ;
;----------------------------------------------------------------------------;
		align	4
public		v64PointB4
proc		v64PointB4	near

pointB4ns:	add	cx,[ds:vCurSRow]
		mov	ax,[ds:vYBytes]
		mul	cx		; point to yStart
		mov	cl,bl		; get bit shift count
		shr	bx,3		; fix X value
		add	bx,ax		; get starting address in BX-reg
		adc	dx,0		; add overflow
		SETBANK			; set to the proper page

		push	ds
		mov	ds,[ds:vSeg]

		and	cl,007h
		mov	al,080h		; set bit mask
		shr	al,cl		; shift to proper position
		SVGAMEM	bx,al		; set point

		pop	ds
		ret
		endp

;----------------------------------------------------------------------------;
; v64HzLineB4								     ;
;									     ;
; Planar horizontal line draw routine using 64k banks.			     ;
;									     ;
; Called with								     ;
;    BX = xStart							     ;
;    CX = yStart							     ;
;    SI = xEnd								     ;
;    DI = yEnd								     ;
;									     ;
; Assumed:								     ;
;    DS = Data Segment.							     ;
;    Line has been properly clipped.					     ;
;    Write mode #3.							     ;
;    Color set in S/R register.						     ;
;									     ;
; Uses:									     ;
;    AX, BX, CX, DX							     ;
;----------------------------------------------------------------------------;
		align	4
public		v64HzLineB4
proc		v64HzLineB4	near

; Test line direction, or single point

		cmp	bx,si
		je	pointB4ns	; if equal, display point
		jb	hplRight	; draw from left to right
		xchg	bx,si		; fix direction

; Save X addresses

hplRight:	add	cx,[ds:vCurSRow]
		add	di,[ds:vCurSRow]

		mov	ax,si		; get xEnd
		mov	dx,bx		; get xStart

; Get starting X-mask in BL-reg

		mov	cl,dl		; get lower 3 bits
		and	cl,007h
		mov	bl,0FFh
		shr	bl,cl		; get mask

; Get ending X-mask in BH-reg

		mov	cl,al		; get lower 3 bits
		and	cl,007h
		mov	bh,080h
		sar	bh,cl		; get mask
		
; calculate number of bytes

		shr	ax,3		; divide by 8
		shr	dx,3		; divide by 8
		sub	ax,dx		; get number of bytes

		mov	cx,ax		; save count

		mov	si,dx		; save xStart
		mov	ax,[ds:vYBytes]
		mul	di		; point to start/end address of Y
		add	si,ax		; add xStart, calc address of vga memory
		adc	dx,0		; add overflow
		SETBANK			; set to the proper starting page

		push	ds
		mov	ds,[ds:vSeg]	; get video memory segment

		dec	cx		; count number of whole bytes
		js	hzr1B		; one byte wide, draw
		jnz	hzrMB		; more than two bytes wide, draw

; Set two bytes

hzr2B:		SVGAMEM	si,bl		; set start mask
		inc	si
		jz	hzr2Bank	; set to new bank
		SVGAMEM	si,bh		; set end mask
		pop	ds
		ret

hzr2Bank:	call	vIncBank
		SVGAMEM	si,bh		; set end mask
		pop	ds
		ret

; Set One byte

		align	4
hzr1B:		and	bl,bh		; and masks
		SVGAMEM	si,bl		; set line segment
		pop	ds
		ret

; More than two bytes wide, draw.

		align	4
hzrMB:		mov	di,ds		; get VGA memory segment in ES & DS
		mov	es,di
		mov	di,si		; get address in ES:DI
		neg	si		; check for count to end of bank
		jz	hzrMB1		; cannot wrap, use as is
		cmp	si,cx		; wrap through bank?
		jae	hzrMB1		; no, use as is
		xchg	cx,si		; use end of bank as count
		sub	si,cx		; get remainder of count in SI-reg
		jmp	Short hzrMB2	; do first bank

hzrMB1:		sub	si,si		; zero remainder count
hzrMB2:		SVGAMEM	di,bl		; set start mask
		inc	di		; point to next

		mov	eax,0FFFFFFFFh	; get mask

; Write bank of data

hzBank:		test	di,0003h	; dword boundry?
		jz	hzDW		; yes, do dwords
		test	di,0001h	; word boundry?
		jz	hzWD		; yes, do words

hzBT:		stos	[Byte es:di]	; make a word boundry
		dec	cx		; done?
		jz	hzDone		; yes, done with bank

hzWD:		dec	cx		; done?
		jz	hzBDone		; yes, write byte, done with bank
		stos	[Word es:di]	; else, just do a full word
		dec	cx		; done?
		jz	hzDone		; yes, exit

		; deal with double words

hzDW:		mov	dl,cl		; get count
		shr	cx,2		; get dword count
		jz	hzDW1		; if no full DWORDs, skip

	rep	stos	[Dword es:di]	; set 32 bit words

hzDW1:		and	dl,003h		; any remainder?
		jz	hzDoneDW	; no, skip
		test	dl,001h		; one word left?
		jz	hzWDone		; yes, set word, done with bank
		dec	dl		; one byte left?
		jz	hzBDone		; yes, set byte, done with bank

		; three bytes left

		stos	[Word es:di]	; set full word
hzBDone:	stos	[Byte es:di]	; set byte
hzDone:		test	si,si		; anymore left?
		jz	hzLast		; no, set last byte

		call	vIncBank	; else, increment bank
		mov	cx,si		; get remainder count
		sub	si,si		; zero remainder
		jmp	Short hzBank	; continue on byte boundry

hzWDone:	stos	[Word es:di]	; set word
		test	si,si		; anymore left?
		jz	hzLast		; set last byte

		call	vIncBank	; else, increment bank
		mov	cx,si		; get remainder count
		sub	si,si		; zero remainder
		jmp	Short hzBank	; continue on word boundry

; Last bank ended on double word boundry, so continue on boundry

hzDoneDW:	test	si,si		; anymore left?
		jz	hzLast		; set last byte

		call	vIncBank	; else, increment bank
		mov	cx,si		; get remainder count
		sub	si,si		; zero remainder
		jmp	Short hzDW	; continue on word boundry

hzLast:		SVGAMEM	di,bh		; set ending mask
		pop	ds
		ret

		endp

;----------------------------------------------------------------------------;
; v64DgLineB4								     ;
;									     ;
; Planar draw a diagonal line using 64k banks.				     ;
;									     ;
; Called with								     ;
;    BX = xStart							     ;
;    CX = yStart							     ;
;    SI = xEnd								     ;
;    DI = yEnd								     ;
;									     ;
; Assumed:								     ;
;    DS = Data Segment.							     ;
;    Line has been properly clipped, and is not horizontal/vertical/point.   ;
;    Write mode #3.							     ;
;    Color set in S/R register.						     ;
;									     ;
;  Line drawing routines using Bresenham's alorithm.			     ;
;----------------------------------------------------------------------------;
		align	4
public		v64DgLineB4
proc		v64DgLineB4	near

		add	cx,[ds:vCurSRow]
		add	di,[ds:vCurSRow]

		mov	ax,si		; xEnd
		sub	ax,bx		; AX = Length of X
		jae	dg1		; if positive, skip
		neg	ax		; else, fix length
		xchg	bx,si		; always draw from left to right
		xchg	cx,di

dg1:		mov	dx,di		; yEnd
		sub	dx,cx		; DX = Length of Y
		jae	dg2		; if positive, skip
		neg	dx

dg2:		cmp	ax,dx		; check which line is longer

		jb	dgB4YXp		; Y is independent axis
;*		jmp	dgB4XXp		; else, X is independent axis
;*
;* Fall through to 'dgB4XXp' ...

		endp

;----------------------------------------------------------------------------;
; dgB4XXp								     ;
;									     ;
; Routine for drawing a line with X as the independent variable.	     ;
; X is moving in a positive (left to right) direction.			     ;
;									     ;
; Called with								     ;
;    AX = xLen								     ;
;    DX = yLen								     ;
;    BX = xStart							     ;
;    CX = yStart							     ;
;    SI = xEnd								     ;
;    DI = yEnd								     ;
;									     ;
; Assumed:								     ;
;    DS = Data Segment.							     ;
;    Line has been properly clipped.					     ;
;    Write mode #3.							     ;
;    Color set in S/R register.						     ;
;----------------------------------------------------------------------------;
		align	4

proc		dgB4XXp		near
		push	bp

		mov	si,[ds:vYBytes] ; get Yinc
		cmp	di,cx		; Y moving positive?
		sbb	bp,bp		; FFFF - if DI > CX
		xor	si,bp		; two's compl if Y moving negative
		sub	si,bp
		
		mov	di,dx		; init neg inc (if D < 0)
		add	di,di		; yLen * 2

		mov	bp,ax		; get xLen
		add	bp,bp		; xLen * 2
		neg	bp
		add	bp,di		; yLen2 - xLen2

		push	ax		; save length

		mov	ax,[ds:vYBytes]
		mul	cx		; point to Y start address
		mov	cl,bl		; get X mask address
		shr	bx,3		; divide by 8 to get byte offset
		add	bx,ax		; point to start memory
		adc	dx,0		; add overflow
		SETBANK			; set to the proper starting page

		and	cl,007h
		mov	dh,080h
		shr	dh,cl		; get initial mask

		pop	cx		; restore length
		mov	ax,si		; get Y increment in AX-reg
		mov	si,di		; init decision variable
		sub	si,cx		; (yLen * 2) - xLen
		inc	cx		; get number of points

		push	ds
		
		mov	ds,[ds:vSeg]	; get video segment

; Register usage:
;
;    DH = Initial mask
;    BX = Starting address
;    CX = Number of points
;    SI = Decision variable
;    BP = Positive incr
;    DI = Negative incr
;    AX = Y increment

		sub	dl,dl		; clear bit mask

		test	ax,ax		; positive increment?
		jns	dgXpPInc	; yes, do positive

;----------------------------------------------------------------------------;
; Y increment is negative						     ;
;----------------------------------------------------------------------------;
		not	ax		; fix increment (negate and -1)
		
		test	si,si		; is value negative?
		js	NdgXpNeg	; yes, start with negative routine

		align	4
NdgXpPos:	or	dl,dh		; add next bit to mask
		dec	cx		; any more points?
		jz	NdgXRet		; no, return

; Positive decision variable - write new mask, increment Y

NdgXp1:		SVGAMEM	bx,dl		; write mask to memory
		sub	dl,dl		; clear bit mask
		ror	dh,1		; rotate mask bit right
		cmc			; compliment carry
		sbb	bx,ax		; decrement
		jc	NdgXp1Bank	; set to new bank
NdgXp1Ret:	add	si,bp		; add positive increment
		jns	NdgXpPos	; do positive decision

		; else do negative decision

		or	dl,dh		; add next bit to mask
		dec	cx		; any more points?
		jz	NdgXRet		; no, return

; Negative decision variable - Increment independent - X

NdgXp2:		ror	dh,1		; rotate mask bit right
		jc	NdgXp4		; if wrap, draw point
		add	si,di		; add negative increment
		jns	NdgXpPos	; do positive decision

		; Else fall through to negative decision

		or	dl,dh		; add next bit to mask
		dec	cx		; any more points?
		jnz	NdgXp2		; yes, draw them
		SVGAMEM	bx,dl		; write mask to memory
		pop	ds
		pop	bp
		ret

; Draw point incase of wrap to next byte

NdgXp4:		SVGAMEM	bx,dl		; write mask to memory
		sub	dl,dl		; clear bit mask
		inc	bx		; point to next byte
		jz	NdgXp4Bank	; set to next bank
NdgXp4Ret:	add	si,di		; add negative increment
		jns	NdgXpPos	; do positive decision

		; Else fall through to negative decision

NdgXpNeg:	or	dl,dh		; add next bit to mask
		dec	cx		; any more points?
		jnz	NdgXp2		; yes, draw them

NdgXRet:	SVGAMEM	bx,dl		; write mask to memory
		pop	ds
		pop	bp
		ret

NdgXp1Bank:	call	vDecBank
		jmp	NdgXp1Ret

NdgXp4Bank:	call	vIncBank
		jmp	NdgXp4Ret

;----------------------------------------------------------------------------;
; Y increment is positive						     ;
;----------------------------------------------------------------------------;

dgXpPInc:	test	si,si		; is value negative?
		js	dgXpNeg		; yes, start with negative routine

		align	4
dgXpPos:	or	dl,dh		; add next bit to mask
		dec	cx		; any more points?
		jz	dgXRet		; no, return

; Positive decision variable - write new mask, increment Y

dgXp1:		SVGAMEM	bx,dl		; write mask to memory
		sub	dl,dl		; clear bit mask
		ror	dh,1		; rotate mask bit right
		adc	bx,ax		; inc and carry
		jc	dgXp1Bank	; set to new bank
dgXp1Ret:	add	si,bp		; add positive increment
		jns	dgXpPos		; do positive decision

		; else do negative decision

		or	dl,dh		; add next bit to mask
		dec	cx		; any more points?
		jz	dgXRet		; no, return

; Negative decision variable - Increment independent - X

dgXp2:		ror	dh,1		; rotate mask bit right
		jc	dgXp4		; if wrap, draw point
		add	si,di		; add negative increment
		jns	dgXpPos		; do positive decision

		; Else fall through to negative decision

		or	dl,dh		; add next bit to mask
		dec	cx		; any more points?
		jnz	dgXp2		; yes, draw them
		SVGAMEM	bx,dl		; write mask to memory
		pop	ds
		pop	bp
		ret

; Draw point incase of wrap to next byte

dgXp4:		SVGAMEM	bx,dl		; write mask to memory
		sub	dl,dl		; clear bit mask
		inc	bx		; point to next byte
		jz	dgXp4Bank	; set to next bank
dgXp4Ret:	add	si,di		; add negative increment
		jns	dgXpPos		; do positive decision

		; Else fall through to negative decision

dgXpNeg:	or	dl,dh		; add next bit to mask
		dec	cx		; any more points?
		jnz	dgXp2		; yes, draw them

dgXRet:		SVGAMEM	bx,dl		; write mask to memory
		pop	ds
		pop	bp
		ret

dgXp1Bank:	call	vIncBank
		jmp	dgXp1Ret

dgXp4Bank:	call	vIncBank
		jmp	dgXp4Ret

		endp

;----------------------------------------------------------------------------;
; dgB4YXp								     ;
;									     ;
; Routine for drawing a line with Y as the independent variable.	     ;
; X is moving in a positive (left to right) direction.			     ;
;									     ;
; Called with								     ;
;    AX = xLen								     ;
;    DX = yLen								     ;
;    BX = xStart							     ;
;    CX = yStart							     ;
;    SI = xEnd								     ;
;    DI = yEnd								     ;
;									     ;
; Assumed:								     ;
;    DS = Data Segment.							     ;
;    Line has been properly clipped.					     ;
;    Write mode #3.							     ;
;    Color set in S/R register.						     ;
;----------------------------------------------------------------------------;
		align	4
proc		dgB4YXp		near
		push	bp

		mov	si,[ds:vYBytes]	; get Yinc
		cmp	di,cx		; Y moving positive?
		sbb	bp,bp		; get 0000 or FFFF - if neg
		xor	si,bp		; two's compl if Y moving negative
		sub	si,bp

		mov	di,ax		; init neg inc (if D < 0)
		add	di,di		; xLen * 2

		mov	bp,dx		; get yLen
		add	bp,bp		; (yLen * 2)
		neg	bp		; bp = di - bp
		add	bp,di		; pre add correction

		push	dx		; save length

		mov	ax,[ds:vYBytes]
		mul	cx		; point to Y start address
		mov	cl,bl		; get X mask address
		shr	bx,3		; divide by 8 to get byte offset
		add	bx,ax		; point to start memory
		adc	dx,0		; add overflow
		SETBANK			; set to the proper starting page

		and	cl,007h
		mov	dh,080h
		shr	dh,cl		; get initial mask

		pop	cx		; restore length
		mov	ax,di		; init decision variable
		sub	ax,cx		; (xLen * 2) - yLen
		inc	cx		; get number of points

		push	ds
		mov	ds,[ds:vSeg]	; get video segment

; Register usage:
;
;    DH = Initial mask
;    BX = Starting address
;    CX = Number of points
;    AX = Decision variable
;    BP = Position increment
;    DI = Negative increment
;    SI = Y increment

		test	si,si		; is Y increment positive?
		jns	dgYpPInc	; yes, do positive Y movement

;----------------------------------------------------------------------------;
; Y increment is positive						     ;
;----------------------------------------------------------------------------;
		not	si		; fix increment (negate and -1)
		
		test	ax,ax		; is value negative?
		js	NdgYpNeg	; yes, start with negative routine

		align	4
NdgYpPos:	SVGAMEM	bx,dh		; write mask to memory
		dec	cx		; any more points?
		jz	NdgYRet		; no, return

; Routine for Positive decision variable - Increment X 

NdgYp1:		ror	dh,1		; rotate mask bit right
		cmc			; compliment carry
		sbb	bx,si		; dec Y with carry
		jc	NdgYp1Bank	; if bank change, do it
NdgYp1Ret:	add	ax,bp		; sub positive increment
		jns	NdgYpPos	; if greater than zero, jump

		SVGAMEM	bx,dh		; write mask to memory
		dec	cx		; any more points?
		jz	NdgYRet		; no, done

; Routine for Negative decision variable - Adjust independent - Y

NdgYp2:		stc			; set c-flg
		sbb	bx,si		; decrement Y
		jc	NdgYp2Bank	; if bank change, do it
NdgYp2Ret:	add	ax,di		; add negative increment
		jns	NdgYpPos	; if greater than zero, jump

		; else, fall through to less than zero routine

NdgYpNeg:	SVGAMEM	bx,dh		; write mask to memory
		dec	cx		; any more points?
		jnz	NdgYp2		; yes, draw them

NdgYRet:	pop	ds
		pop	bp
		ret

NdgYp1Bank:	call	vDecBank
		jmp	NdgYp1Ret

NdgYp2Bank:	call	vDecBank
		jmp	NdgYp2Ret

;----------------------------------------------------------------------------;
; Y increment is positive						     ;
;----------------------------------------------------------------------------;
dgYpPInc:	test	ax,ax		; is value negative?
		js	dgYpNeg		; yes, start with negative routine

		align	4
dgYpPos:	SVGAMEM	bx,dh		; write mask to memory
		dec	cx		; any more points?
		jz	dgYRet		; no, return

; Routine for Positive decision variable - Increment X 

dgYp1:		ror	dh,1		; rotate mask bit right
		adc	bx,si		; add Y increment, and carry
		jc	dgYp1Bank	; if bank change, do it
dgYp1Ret:	add	ax,bp		; sub positive increment
		jns	dgYpPos		; if greater than zero, jump

		SVGAMEM	bx,dh		; write mask to memory
		dec	cx		; any more points?
		jz	dgYRet		; no, done

; Routine for Negative decision variable - Adjust independent - Y

dgYp2:		add	bx,si		; add Y increment
		jc	dgYp2Bank	; if bank change, do it
dgYp2Ret:	add	ax,di		; add negative increment
		jns	dgYpPos		; if greater than zero, jump

		; else, fall through to less than zero routine

dgYpNeg:	SVGAMEM	bx,dh		; write mask to memory
		dec	cx		; any more points?
		jnz	dgYp2		; yes, draw them

dgYRet:		pop	ds
		pop	bp
		ret

dgYp1Bank:	call	vIncBank
		jmp	dgYp1Ret

dgYp2Bank:	call	vIncBank
		jmp	dgYp2Ret

		endp

		ends
		end
