;----------------------------------------------------------------------------;
; Misc routines for a VGA planar 4 bit (16 color) display.
;
; Author: Zonn Moore
; Date:   3/16/00
;
; (c) Copyright 2000, Zonn Moore.  All rights reserved.
;----------------------------------------------------------------------------;
		ideal
		P386N
		include	"zvga.inc"
		include	"vesa.inc"
		include	"v64B4.inc"

segment		_DATA	public 'DATA' dword
		extrn	_vgSeg:Word,vgVecPtr:Dword

		public	vSeg,vWinSize,vXSize,vYSize,vYBytes
		public	vCurSRow,vLastSRow,vCurBank

		public	vSetup,vXltBi,vXlt16,vXlt64,vXltColor,vSetColor
		public	vSetErase,vSetDraw,vCls,vPoint,vLine,vBriColor

; Table of vector routines called

vSetup		dw	vSetupB4	; setup before writing
vXltBi		dw	vXltBiB4	; translate a bi-level color
vXlt16		dw	vXlt16B4	; translate a 16-level color
vXlt64		dw	vXlt64B4	; translate a 64-level color
vXltColor	dw	vXltColorB4	; translate a 4096-level color
vSetColor	dw	vSetColorB4	; set draw color
vSetErase	dw	vSetEraseB4	; setup to erase
vSetDraw	dw	vSetDrawB4	; setup to draw
vCls		dw	v64ClsB4	; clear screen
vPoint		dw	v64PointB4	; draw a point
vLine		dw	v64LineB4	; draw a line

vBriColor	dw	0000Fh		; bright color
vSeg		dw	?		; segment of video memory
vWinSize	dd	?		; size of window in bytes
vXSize		dw	?		; number of pixels in X dir
vYSize		dw	?		; number of pixels in Y dir
vYBytes		dw	?		; number of bytes in Y dir

vCurSRow	dw	?		; current starting row
vLastSRow	dw	?		; last pages starting row

vCurBank	dw	?		; current bank of writable memory

; Vesa structures

vbeInfo		VbeInfoBlock_s		?
vbeMode		ModeInfoBlock_s		?

		ends

segment		_TEXT	public 'CODE' dword
		assume	cs:_TEXT

;----------------------------------------------------------------------------;
; Paging routines needed for crossing 64k boundries.			     ;
;----------------------------------------------------------------------------;

;----------------------------------------------------------------------------;
; vInit									     ;
;----------------------------------------------------------------------------;
public		vInit
proc		vInit

 if 0

; Vesa test code

		; get information on the VESA drivers

		mov	di,ds			; get data segment
		mov	es,di
		mov	di,Offset vbeInfo	; point to structure
		mov	[Dword es:di+VbeInfoBlock_s.VbeSignature],'2EBV'
		mov	ax,vesaGetCtrlInfo
		int	10h			; call video BIOS

		cmp	al,04Fh			; VESA present?
		jne	vInitVGA		; no, use VGA mode 10h
		cmp	[Dword es:di+VbeInfoBlock_s.VbeSignature],'ASEV'
		jne	vInitVGA		; if no match, use VGA mode 10h
		cmp	[es:di+VbeInfoBlock_s.VbeVersion],0200h ; check for invalid VESA version
		ja	vesa1			; set to VESA 1.2
		cmp	[es:di+VbeInfoBlock_s.OemVendorNamePtr],0
		jne	vesa2			; if not 0, skip
vesa1:		mov	[es:di+VbeInfoBlock_s.VbeVersion],0102h ; set to VESA 1.2

vesa2:		mov	ax,4F02h		; set mode
		mov	bx,102h
		int	10h
		cmp	ax,004Fh
		jne	vInitVGA

		mov	cx,102h
		mov	ax,4F01H
		mov	di,Offset vbeMode
		int	10h
		cmp	ax,004Fh
		jne	vInitVGA

 if 1
		mov	ax,[ds:vbeMode.WinASegment]
		mov	[ds:vSeg],ax
		mov	[ds:vXSize],800
		mov	[ds:vYSize],600
		mov	[ds:vYBytes],800 / 8
		mov	[ds:vWinSize],(800 / 8) * 600

		mov	[ds:vCurSRow],0
		mov	[ds:vLastSRow],600
 else
		mov	[ds:vSeg],VGASEG1
		mov	[ds:vXSize],VGAX
		mov	[ds:vYSize],VGAY
		mov	[ds:vYBytes],VGAX / 8
		mov	[ds:vWinSize],VGASIZE

		mov	[ds:vCurSRow],0
		mov	[ds:vLastSRow],VGAY
 endif

		mov	[ds:vCurBank],0		; set current bank to zero
		call	vSetupB4
		ret
 endif

vInitVGA:	mov	al,MODE		; get mode
		mov	ah,0		; set mode
		int	10h		; call video BIOS

		mov	[ds:vSeg],VGASEG1
		mov	[ds:vXSize],VGAX
		mov	[ds:vYSize],VGAY
		mov	[ds:vYBytes],VGAX / 8
		mov	[ds:vWinSize],VGASIZE

		mov	[ds:vCurSRow],0
		mov	[ds:vLastSRow],VGAY

		mov	[ds:vCurBank],-1 ; set so current bank never matches

 if 0

; Setup test code for ATI

		mov	dx,01CEh
		mov	al,0B0h
		out	dx,al
		inc	dx
		in	al,dx
		dec	dx
		mov	ah,al
		mov	al,0B0h
		out	dx,al
		inc	dx
		mov	al,ah
		and	al,0C7h
		or	al,010h
		out	dx,al

		dec	dx
		mov	al,0B6h
		out	dx,al
		inc	dx
		in	al,dx
		dec	dx
		mov	ah,al
		mov	al,0B6h
		out	dx,al
		inc	dx
		mov	al,ah
		or	al,001h
		out	dx,al

		dec	dx
		mov	al,0BFh
		out	dx,al
		inc	dx
		in	al,dx
		dec	dx
		mov	ah,al
		mov	al,0BFh
		out	dx,al
		inc	dx
		mov	al,ah
;		or	al,020h			; vertical ret changes
		and	al,0DFh			; vertical ret changes
		out	dx,al
 endif

		call	[ds:vSetup]
		call	[ds:vSetErase]
		call	[ds:vCls]
		ret
		endp

public		vSetBank
proc		vSetBank

 if 0

; Test code for ATI

		push	dx
		push	ax
		mov	[ds:vCurBank],dx
		and	dl,007h			; get only page (0-7)
		shl	dl,1
		mov	ah,dl			; get page
		shl	dl,4
		or	ah,dl
		mov	dx,01CEh
		mov	al,0B2h
		cli
		out	dx,al
		inc	dx
		mov	al,ah
		out	dx,al
		sti
		pop	ax
		pop	dx
 endif
		ret
		endp

public		vIncBank
proc		vIncBank

 if 0

; Test code for ATI

		push	dx
		push	ax
		inc	[gs:vCurBank]
		mov	dl,[Byte gs:vCurBank]
		and	dl,007h			; get only page (0-7)
		shl	dl,1
		mov	ah,dl			; get page
		shl	dl,4
		or	ah,dl
		mov	dx,01CEh
		mov	al,0B2h
		cli
		out	dx,al
		inc	dx
		mov	al,ah
		out	dx,al
		sti
		pop	ax
		pop	dx
 endif

		ret
		endp

public		vDecBank
proc		vDecBank

 if 0

; Test code for ATI

		push	dx
		push	ax
		dec	[gs:vCurBank]
		mov	dl,[Byte gs:vCurBank]
		and	dl,007h			; get only page (0-7)
		shl	dl,1
		mov	ah,dl			; get page
		shl	dl,4
		or	ah,dl
		mov	dx,01CEh
		mov	al,0B2h
		cli
		out	dx,al
		inc	dx
		mov	al,ah
		out	dx,al
		sti
		pop	ax
		pop	dx
 endif
		ret
		endp

;----------------------------------------------------------------------------;
; Swap buffers.								     ;
;									     ;
; Called with:								     ;
;    DS = Pointer to DataSegment					     ;
;----------------------------------------------------------------------------;
public		vSwap
proc		vSwap

		mov	ax,[ds:vCurSRow]	; get current starting row
		mul	[ds:vYBytes]		; mult by bytes per row
		mov	bx,ax			; starting adress in CX:BX
		mov	cx,dx

 if 0
		cmp	[Word Ptr ds:vWinSize+2],0
		jz	vsw1

		shl	cl,6
		and	cl,040h
		mov	dx,01CEh
		mov	al,0B0h
		cli

		out	dx,al
		inc	dx
		in	al,dx
		dec	dx
		mov	ah,al
		mov	al,0B0h
		out	dx,al
		inc	dx
		mov	al,ah
		and	al,0BFh
		or	al,cl
		out	dx,al
 endif

vsw1:		mov	dx,VGACRT		; point to CRT registers
		mov	al,crtSAH		; Misc Register
		cli
		out	dx,al			; point to high reg
		inc	dx
		mov	al,bh
		out	dx,al			; display new page
		dec	dx
		mov	al,crtSAL		; point to low reg
		out	dx,al
		inc	dx
		mov	al,bl
		out	dx,al
		sti


 if 0
		mov	ax,04F07h		; set start of display
		mov	bx,0h			; set display
		sub	cx,cx			; start at X of zero
		mov	dx,[ds:vCurSRow]	; new starting Y
;		call	[ds:vbeMode.WinFuncPtr]
 endif

		mov	ax,[ds:vCurSRow]
		xchg	ax,[ds:vLastSRow]
		mov	[ds:vCurSRow],ax

		ret
		endp

;----------------------------------------------------------------------------;
; calcInc								     ;
;									     ;
; Called with:								     ;
;   AX = Highest value							     ;
;   DX = Lowest value							     ;
;   CX = Number of intervals						     ;
;									     ;
; Returns with:								     ;
;   DX:AX as Integer and Fraction of increment needed to step through	     ;
;   all values.								     ;
;----------------------------------------------------------------------------;
public		calcInc
proc		calcInc
		cmp	ax,dx			; is AX < DX
		jb	negInc			; yes, then negative inc

posInc:		sub	ax,dx			; get range
		sub	dx,dx			; zero upper word
		div	cx			; divide by intervals
		push	ax			; save integer part
		sub	ax,ax			; zero lower half
		div	cx			; get fractional part
		pop	dx			; restore integer part
		ret

negInc:		xchg	ax,dx			; fix order
		call	posInc			; get increment
		neg	dx			; negate
		neg	ax			; negate
		sbb	dx,0			; fix if underflow
		ret
		endp

		ends
		end
